import numpy as np
from scipy.constants import epsilon_0
from geoana.utils import check_xyz_dim

class ElectrostaticSphere:
    """Class for electrostatic solutions for a sphere in a wholespace.

    The ``ElectrostaticSphere`` class is used to analytically compute the electric
    potentials, fields, currents and charge densities for a sphere in a wholespace.

    Parameters
    ----------
    radius : float
        radius of sphere (m).
    sigma_sphere : float
        conductivity of target sphere (S/m).
    sigma_background : float
        background conductivity (S/m).
    primary_field : (3) array_like, optional
        amplitude of primary electric field.  Default is (1, 0, 0).
    location : (3) array_like, optional
        center of the sphere. Default is (0, 0, 0).
    """

    def __init__(
        self, radius, sigma_sphere, sigma_background, primary_field=None, location=None
    ):

        self.radius = radius
        self.sigma_sphere = sigma_sphere
        self.sigma_background = sigma_background
        if primary_field is None:
            primary_field = np.r_[1., 0., 0.]
        self.primary_field = primary_field
        if location is None:
            location = np.r_[0., 0., 0.]
        self.location = location

    @property
    def sigma_sphere(self):
        """Electrical conductivity of the sphere in S/m.

        Returns
        -------
        float
            Electrical conductivity of the sphere in S/m.
        """
        return self._sigma_sphere

    @sigma_sphere.setter
    def sigma_sphere(self, item):
        item = float(item)
        if item <= 0.0:
            raise ValueError('Conductiviy must be positive')
        self._sigma_sphere = item

    @property
    def sigma_background(self):
        """Electrical conductivity of the background in S/m.

        Returns
        -------
        float
            Electrical conductivity of the background in S/m.
        """
        return self._sigma_background

    @sigma_background.setter
    def sigma_background(self, item):
        item = float(item)
        if item <= 0.0:
            raise ValueError('Conductiviy must be positive')
        self._sigma_background = item

    @property
    def radius(self):
        """Radius of the sphere in meters.

        Returns
        -------
        float
            Radius of the sphere in meters.
        """
        return self._radius

    @radius.setter
    def radius(self, item):
        item = float(item)
        if item < 0.0:
            raise ValueError('radius must be non-negative')
        self._radius = item

    @property
    def primary_field(self):
        """Amplitude of the primary current density.

        Returns
        -------
        (3) numpy.ndarray of float
            Amplitude of the primary current density. Default = np.r_[1,0,0]
        """
        return self._primary_field

    @primary_field.setter
    def primary_field(self, vec):
        try:
            vec = np.atleast_1d(vec).astype(float)
        except:
            raise TypeError(f"primary_field must be array_like, got {type(vec)}")

        if len(vec) != 3:
            raise ValueError(
                f"primary_field must be array_like with shape (3,), got {len(vec)}"
            )

        self._primary_field = vec

    @property
    def location(self):
        """Center of the sphere.

        Returns
        -------
        (3) numpy.ndarray of float
            Center of the sphere. Default = np.r_[0,0,0]
        """
        return self._location

    @location.setter
    def location(self, vec):

        try:
            vec = np.atleast_1d(vec).astype(float)
        except:
            raise TypeError(f"location must be array_like, got {type(vec)}")

        if len(vec) != 3:
            raise ValueError(
                f"location must be array_like with shape (3,), got {len(vec)}"
            )

        self._location = vec

    def potential(self, xyz, field='all'):
        """Electric potential for a conductive sphere in a uniform electrostatic field.

       .. math::

            V_p(\\mathbf{r}) = -\\mathbf{E_0} \\dot \\mathbf{r}

            r > R

            V_T(\\mathbf{r}) = -\\mathbf{E_0} \\dot \\mathbf{r} (1 - \\frac{\\sigma_s - \\sigma_0}{\\sigma_s + 2 \\sigma_0} \\frac{R}{r^3}

            r < R

            V_T(\\mathbf{r}) = -3 \\mathbf{E_0} \\dot \\mathbf{r} \\frac{\\sigma_0}{\\sigma_s + 2 \\sigma_0}

        Parameters
        ----------
        xyz : (..., 3) numpy.ndarray
            Locations to evaluate at in units m.
        field : {'all', 'total', 'primary', 'secondary'}

        Returns
        -------
        Vt, Vp, Vs : (..., ) np.ndarray
            If field == "all"
        V : (..., ) np.ndarray
            If only requesting a single field.

        Examples
        --------
        Here, we define a sphere with conductivity sigma_sphere in a uniform electrostatic field with conductivity
        sigma_background and plot the total and secondary electric potentials.

        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from matplotlib import patches
        >>> from mpl_toolkits.axes_grid1 import make_axes_locatable
        >>> from geoana.em.static import ElectrostaticSphere

        Define the sphere.

        >>> sigma_sphere = 10. ** -1
        >>> sigma_background = 10. ** -3
        >>> radius = 1.0
        >>> simulation = ElectrostaticSphere(
        >>>     location=None, sigma_sphere=sigma_sphere, sigma_background=sigma_background, radius=radius, primary_field=None
        >>> )

        Now we create a set of gridded locations and compute the magnetic potential.

        >>> X, Y = np.meshgrid(np.linspace(-2*radius, 2*radius, 20), np.linspace(-2*radius, 2*radius, 20))
        >>> Z = np.zeros_like(X) + 0.25
        >>> xyz = np.stack((X, Y, Z), axis=-1)
        >>> vt = simulation.potential(xyz, field='total')
        >>> vs = simulation.potential(xyz, field='secondary')

        Finally, we plot the total and secondary electric potentials.

        >>> fig, axs = plt.subplots(1, 2, figsize=(18,12))
        >>> titles = ['Total Potential', 'Secondary Potential']
        >>> for ax, V, title in zip(axs.flatten(), [vt, vs], titles):
        >>>     im = ax.pcolor(X, Y, V, shading='auto')
        >>>     divider = make_axes_locatable(ax)
        >>>     cax = divider.append_axes("right", size="5%", pad=0.05)
        >>>     cb = plt.colorbar(im, cax=cax)
        >>>     cb.set_label(label= 'Potential (V)')
        >>>     ax.add_patch(patches.Circle((0, 0), radius, fill=False, linestyle='--'))
        >>>     ax.set_ylabel('Y coordinate ($m$)')
        >>>     ax.set_xlabel('X coordinate ($m$)')
        >>>     ax.set_title(title)
        >>>     ax.set_aspect('equal')
        >>> plt.tight_layout()
        >>> plt.show()
        """
        xyz = check_xyz_dim(xyz)
        sig0 = self.sigma_background
        sig1 = self.sigma_sphere
        E0 = self.primary_field
        sig_cur = (sig1 - sig0) / (sig1 + 2 * sig0)
        r_vec = xyz - self.location
        r = np.linalg.norm(r_vec, axis=-1)

        Vt = np.zeros_like(r)
        ind0 = r > self.radius

        # total potential outside the sphere
        Vt[ind0] = r_vec[ind0] @ -E0 * (1. - sig_cur * self.radius ** 3. / r[ind0] ** 3.)

        # inside the sphere
        Vt[~ind0] = r_vec[~ind0] @ -E0 * 3. * sig0 / (sig1 + 2. * sig0)

        # total field
        if field == 'total':
            return Vt

        # primary field
        if field != 'total':
            Vp = r_vec @ -E0
            if field == 'primary':
                return Vp

        # secondary field
        Vs = Vt - Vp
        if field == 'secondary':
            return Vs

        return Vt, Vp, Vs

    def electric_field(self, xyz, field='all'):
        """Electric field for a sphere in a uniform wholespace

        .. math::

            E_p(\\mathbf{r}) = - \\nabla V_p = \\mathbf{E_0}

        Parameters
        ----------
        xyz : (..., 3) numpy.ndarray
            Locations to evaluate at in units m.
        field : {'all', 'total', 'primary', 'secondary'}

        Returns
        -------
        Et, Ep, Es : (..., 3) np.ndarray
            If field == "all"
        E : (..., 3) np.ndarray
            If only requesting a single field.

        Examples
        --------
        Here, we define a sphere with conductivity sigma_sphere in a uniform electrostatic field with conductivity
        sigma_background and plot the total and secondary electric fields.

        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from matplotlib import patches
        >>> from mpl_toolkits.axes_grid1 import make_axes_locatable
        >>> from geoana.em.static import ElectrostaticSphere

        Define the sphere.

        >>> sigma_sphere = 10. ** -1
        >>> sigma_background = 10. ** -3
        >>> radius = 1.0
        >>> simulation = ElectrostaticSphere(
        >>>     location=None, sigma_sphere=sigma_sphere, sigma_background=sigma_background, radius=radius, primary_field=None
        >>> )

        Now we create a set of gridded locations and compute the electric fields.

        >>> X, Y = np.meshgrid(np.linspace(-2*radius, 2*radius, 20), np.linspace(-2*radius, 2*radius, 20))
        >>> Z = np.zeros_like(X) + 0.25
        >>> xyz = np.stack((X, Y, Z), axis=-1)
        >>> et = simulation.electric_field(xyz, field='total')
        >>> es = simulation.electric_field(xyz, field='secondary')

        Finally, we plot the total and secondary electric fields.

        >>> fig, axs = plt.subplots(1, 2, figsize=(18,12))
        >>> titles = ['Total Electric Field', 'Secondary Electric Field']
        >>> for ax, E, title in zip(axs.flatten(), [et, es], titles):
        >>>     E_amp = np.linalg.norm(E, axis=-1)
        >>>     im = ax.pcolor(X, Y, E_amp, shading='auto')
        >>>     divider = make_axes_locatable(ax)
        >>>     cax = divider.append_axes("right", size="5%", pad=0.05)
        >>>     cb = plt.colorbar(im, cax=cax)
        >>>     cb.set_label(label= 'Amplitude ($V/m$)')
        >>>     ax.streamplot(X, Y, E[..., 0], E[..., 1], density=0.75)
        >>>     ax.add_patch(patches.Circle((0, 0), radius, fill=False, linestyle='--'))
        >>>     ax.set_ylabel('Y coordinate ($m$)')
        >>>     ax.set_xlabel('X coordinate ($m$)')
        >>>     ax.set_aspect('equal')
        >>>     ax.set_title(title)
        >>> plt.tight_layout()
        >>> plt.show()
        """
        xyz = check_xyz_dim(xyz)
        sig0 = self.sigma_background
        sig1 = self.sigma_sphere
        E0 = self.primary_field
        sig_cur = (sig1 - sig0) / (sig1 + 2 * sig0)
        r_vec = xyz - self.location
        r = np.linalg.norm(r_vec, axis=-1)

        Et = np.zeros((*r.shape, 3))
        ind0 = r > self.radius

        # total field outside the sphere
        Et[ind0] = E0 * (1. - sig_cur * self.radius ** 3. / r[ind0, None] ** 3) +\
            3. * (r_vec[ind0] @ E0)[..., None] * sig_cur * self.radius * r_vec[ind0] / r[ind0, None] ** 4

        # inside the sphere
        Et[~ind0] = 3. * sig0 / (sig1 + 2. * sig0) * E0

        # total field
        if field == 'total':
            return Et

        # primary field
        if field != 'total':
            Ep = np.zeros((*r.shape, 3))
            Ep = Ep + E0
            if field == 'primary':
                return Ep

        # secondary field
        Es = Et - Ep
        if field == 'secondary':
            return Es

        return Et, Ep, Es

    def current_density(self, xyz, field='all'):
        """Current density for a sphere in a uniform wholespace

        Parameters
        ----------
        xyz : (..., 3) numpy.ndarray
            Locations to evaluate at in units m.
        field : {'all', 'total', 'primary', 'secondary'}

        Returns
        -------
        Jt, Jp, Js : (..., 3) np.ndarray
            If field == "all"
        J : (..., 3) np.ndarray
            If only requesting a single field.

        Examples
        --------
        Here, we define a sphere with conductivity sigma_sphere in a uniform electrostatic field with conductivity
        sigma_background and plot the total and secondary current densities.

        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from matplotlib import patches
        >>> from mpl_toolkits.axes_grid1 import make_axes_locatable
        >>> from geoana.em.static import ElectrostaticSphere

        Define the sphere.

        >>> sigma_sphere = 10. ** -1
        >>> sigma_background = 10. ** -3
        >>> radius = 1.0
        >>> simulation = ElectrostaticSphere(
        >>>     location=None, sigma_sphere=sigma_sphere, sigma_background=sigma_background, radius=radius, primary_field=None
        >>> )

        Now we create a set of gridded locations and compute the current densities.

        >>> X, Y = np.meshgrid(np.linspace(-2*radius, 2*radius, 20), np.linspace(-2*radius, 2*radius, 20))
        >>> Z = np.zeros_like(X) + 0.25
        >>> xyz = np.stack((X, Y, Z), axis=-1)
        >>> jt = simulation.current_density(xyz, field='total')
        >>> js = simulation.current_density(xyz, field='secondary')

        Finally, we plot the total and secondary current densities.

        >>> fig, axs = plt.subplots(1, 2, figsize=(18,12))
        >>> titles = ['Total Current Density', 'Secondary Current Density']
        >>> for ax, J, title in zip(axs.flatten(), [jt, js], titles):
        >>>     J_amp = np.linalg.norm(J, axis=-1)
        >>>     im = ax.pcolor(X, Y, J_amp, shading='auto')
        >>>     divider = make_axes_locatable(ax)
        >>>     cax = divider.append_axes("right", size="5%", pad=0.05)
        >>>     cb = plt.colorbar(im, cax=cax)
        >>>     cb.set_label(label= 'Current Density ($A/m^2$)')
        >>>     ax.streamplot(X, Y, J[..., 0], J[..., 1], density=0.75)
        >>>     ax.add_patch(patches.Circle((0, 0), radius, fill=False, linestyle='--'))
        >>>     ax.set_ylabel('Y coordinate ($m$)')
        >>>     ax.set_xlabel('X coordinate ($m$)')
        >>>     ax.set_aspect('equal')
        >>>     ax.set_title(title)
        >>> plt.tight_layout()
        >>> plt.show()
        """
        xyz = check_xyz_dim(xyz)
        r = np.linalg.norm(xyz - self.location, axis=-1)

        sigma = np.full(r.shape, self.sigma_background)
        sigma[r <= self.radius] = self.sigma_sphere

        Et = self.electric_field(xyz, field='total')
        Jt = sigma[..., None] * Et
        if field == 'total':
            return Jt

        if field != 'total':
            Ep = self.electric_field(xyz, field='primary')
            Jp = self.sigma_background * Ep
            if field == 'primary':
                return Jp

        Js = Jt - Jp
        if field == 'secondary':
            return Js

        return Jt, Jp, Js

    def charge_density(self, xyz, dr=None):
        """charge density on the surface of a sphere in a uniform wholespace

        Parameters
        ----------
        xyz : (..., 3) numpy.ndarray
            Locations to evaluate at in units m.
        dr : float, optional
            Buffer around the edge of the sphere to calculate
            current density. Defaults to 5 % of the sphere radius

        Returns
        -------
        rho: (..., ) np.ndarray

        Examples
        --------
        Here, we define a sphere with conductivity sigma_sphere in a uniform electrostatic field with conductivity
        sigma_background and plot the charge density.

        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from matplotlib import patches
        >>> from mpl_toolkits.axes_grid1 import make_axes_locatable
        >>> from geoana.em.static import ElectrostaticSphere

        Define the sphere.

        >>> sigma_sphere = 10. ** -1
        >>> sigma_background = 10. ** -3
        >>> radius = 1.0
        >>> simulation = ElectrostaticSphere(
        >>>     location=None, sigma_sphere=sigma_sphere, sigma_background=sigma_background, radius=radius, primary_field=None
        >>> )

        Now we create a set of gridded locations and compute the charge density.

        >>> X, Y = np.meshgrid(np.linspace(-2*radius, 2*radius, 20), np.linspace(-2*radius, 2*radius, 20))
        >>> Z = np.zeros_like(X) + 0.25
        >>> xyz = np.stack((X, Y, Z), axis=-1)
        >>> q = simulation.charge_density(xyz)

        Finally, we plot the charge density.

        >>> fig, ax = plt.subplots(figsize=(18,6))
        >>> im = ax.pcolor(X, Y, q, shading='auto')
        >>> cb1 = plt.colorbar(im, ax=ax)
        >>> cb1.set_label(label= 'Charge Density ($C/m^2$)')
        >>> ax.add_patch(patches.Circle((0, 0), radius, fill=False, linestyle='--'))
        >>> ax.set_ylabel('Y coordinate ($m$)')
        >>> ax.set_xlabel('X coordinate ($m$)')
        >>> ax.set_title('Charge Accumulation')
        >>> ax.set_aspect('equal')
        >>> plt.show()
        """
        xyz = check_xyz_dim(xyz)

        sig0 = self.sigma_background
        sig1 = self.sigma_sphere
        sig_cur = (sig1 - sig0) / (sig1 + 2 * sig0)
        E0 = self.primary_field
        r_vec = xyz - self.location
        r = np.linalg.norm(r_vec, axis=-1)

        if dr is None:
            dr = 0.05 * self.radius

        ind = (r < self.radius + 0.5 * dr) & (r > self.radius - 0.5 * dr)

        rho = np.zeros((*r.shape, 3))
        rho[ind] = epsilon_0 * 3. * (r_vec[ind] @ E0) * sig_cur / r[ind]

        return rho


class MagnetostaticSphere:
    """Class for magnetostatic solutions for a permeable sphere in a uniform magnetostatic field.

        The ``MagnetostaticSphere`` class is used to analytically compute the magnetic
        potentials, fields, and magnetic flux densities for a permeable sphere in a uniform magnetostatic field.

        Parameters
        ----------
        radius : float
            radius of sphere (m).
        mu_sphere : float
            permeability of target sphere (H/m).
        mu_background : float
            background permeability (H/m).
        primary_field : (3) array_like, optional
            amplitude of primary magnetic field.  Default is (1, 0, 0)
        location : (3) array_like, optional
            Center of the sphere. Default is (0, 0, 0).
        """

    def __init__(
        self, radius, mu_sphere, mu_background, primary_field=None, location=None
    ):

        self.radius = radius
        self.mu_sphere = mu_sphere
        self.mu_background = mu_background
        if primary_field is None:
            primary_field = np.r_[1., 0., 0.]
        self.primary_field = primary_field
        if location is None:
            location = np.r_[0, 0, 0]
        self.location = location

    @property
    def mu_sphere(self):
        """Magnetic permeability of the sphere in H/m.

        Returns
        -------
        float
            Magnetic permeability of the sphere in H/m.
        """
        return self._mu_sphere

    @mu_sphere.setter
    def mu_sphere(self, item):
        item = float(item)
        if item <= 0.0:
            raise ValueError('Permeability must be positive')
        self._mu_sphere = item

    @property
    def mu_background(self):
        """Magnetic permeability of the background in H/m.

        Returns
        -------
        float
            Magnetic permeability of the background in H/m.
        """
        return self._mu_background

    @mu_background.setter
    def mu_background(self, item):
        item = float(item)
        if item <= 0.0:
            raise ValueError('Permeability must be positive')
        self._mu_background = item

    @property
    def radius(self):
        """Radius of the sphere in meters.

        Returns
        -------
        float
            Radius of the sphere in meters.
        """
        return self._radius

    @radius.setter
    def radius(self, item):
        item = float(item)
        if item < 0.0:
            raise ValueError('radius must be non-negative')
        self._radius = item

    @property
    def primary_field(self):
        """Amplitude of the primary current density.

        Returns
        -------
        (3) numpy.ndarray of float
            Amplitude of the primary current density.  Default = np.r_[1,0,0]
        """
        return self._primary_field

    @primary_field.setter
    def primary_field(self, vec):

        try:
            vec = np.atleast_1d(vec).astype(float)
        except:
            raise TypeError(f"primary_field must be array_like, got {type(vec)}")

        if len(vec) != 3:
            raise ValueError(
                f"primary_field must be array_like with shape (3,), got {len(vec)}"
            )

        self._primary_field = vec

    @property
    def location(self):
        """Center of the sphere

        Returns
        -------
        (3) numpy.ndarray of float
            Center of the sphere. Default = np.r_[0,0,0]
        """
        return self._location

    @location.setter
    def location(self, vec):

        try:
            vec = np.atleast_1d(vec).astype(float)
        except:
            raise TypeError(f"location must be array_like, got {type(vec)}")

        if len(vec) != 3:
            raise ValueError(
                f"location must be array_like with shape (3,), got {len(vec)}"
            )

        self._location = vec

    def potential(self, xyz, field='all'):
        """Magnetic potential for a permeable sphere in a uniform magnetostatic field.

        Parameters
        ----------
        xyz : (..., 3) numpy.ndarray
            Locations to evaluate at in units m.
        field : {'all', 'total', 'primary', 'secondary'}

        Returns
        -------
        Vt, Vp, Vs : (..., ) np.ndarray
            If field == "all"
        V : (..., ) np.ndarray
            If only requesting a single field.

        Examples
        --------
        Here, we define a sphere with permeability mu_sphere in a uniform magnetostatic field with permeability
        mu_background and plot the total and secondary magnetic potentials.

        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from matplotlib import patches
        >>> from mpl_toolkits.axes_grid1 import make_axes_locatable
        >>> from geoana.em.static import MagnetostaticSphere

        Define the sphere.

        >>> mu_sphere = 10. ** -1
        >>> mu_background = 10. ** -3
        >>> radius = 1.0
        >>> simulation = MagnetostaticSphere(
        >>>     location=None, mu_sphere=mu_sphere, mu_background=mu_background, radius=radius, primary_field=None
        >>> )

        Now we create a set of gridded locations and compute the magnetic potential.

        >>> X, Y = np.meshgrid(np.linspace(-2*radius, 2*radius, 20), np.linspace(-2*radius, 2*radius, 20))
        >>> Z = np.zeros_like(X) + 0.25
        >>> xyz = np.stack((X, Y, Z), axis=-1)
        >>> vt = simulation.potential(xyz, field='total')
        >>> vs = simulation.potential(xyz, field='secondary')

        Finally, we plot the total and secondary magnetic potentials.

        >>> fig, axs = plt.subplots(1, 2, figsize=(18,12))
        >>> titles = ['Total Potential', 'Secondary Potential']
        >>> for ax, V, title in zip(axs.flatten(), [vt, vs], titles):
        >>>     im = ax.pcolor(X, Y, V, shading='auto')
        >>>     divider = make_axes_locatable(ax)
        >>>     cax = divider.append_axes("right", size="5%", pad=0.05)
        >>>     cb = plt.colorbar(im, cax=cax)
        >>>     cb.set_label(label= 'Potential (A)')
        >>>     ax.add_patch(patches.Circle((0, 0), radius, fill=False, linestyle='--'))
        >>>     ax.set_ylabel('Y coordinate ($m$)')
        >>>     ax.set_xlabel('X coordinate ($m$)')
        >>>     ax.set_title(title)
        >>>     ax.set_aspect('equal')
        >>> plt.tight_layout()
        >>> plt.show()
        """
        xyz = check_xyz_dim(xyz)
        mu0 = self.mu_background
        mu1 = self.mu_sphere
        H0 = self.primary_field
        mu_cur = (mu1 - mu0) / (mu1 + 2 * mu0)
        r_vec = xyz - self.location
        r = np.linalg.norm(r_vec, axis=-1)

        Vt = np.zeros_like(r)
        ind0 = r > self.radius

        # total potential outside the sphere
        Vt[ind0] = r_vec[ind0] @ -H0 * (1. - mu_cur * self.radius ** 3. / r[ind0] ** 3.)

        # inside the sphere
        Vt[~ind0] = r_vec[~ind0] @ -H0 * 3. * mu0 / (mu1 + 2. * mu0)

        # total field
        if field == 'total':
            return Vt

        # primary field
        if field != 'total':
            Vp = r_vec @ -H0
            if field == 'primary':
                return Vp

        # secondary field
        Vs = Vt - Vp
        if field == 'secondary':
            return Vs

        return Vt, Vp, Vs

    def magnetic_field(self, xyz, field='all'):
        """Magnetic field for a permeable sphere in a uniform magnetostatic field.  See Ward and Hohmann,
        1988 Equation 6.69.

        .. math::

            \\mathbf{H} = - \\nabla U

        Parameters
        ----------
        xyz : (..., 3) numpy.ndarray
            Locations to evaluate at in units m.
        field : {'all', 'total', 'primary', 'secondary'}

        Returns
        -------
        Ht, Hp, Hs : (..., 3) np.ndarray
            If field == "all"
        H : (..., 3) np.ndarray
            If only requesting a single field.

        Examples
        --------
        Here, we define a sphere with permeability mu_sphere in a uniform magnetostatic field with permeability
        mu_background and plot the total and secondary magnetic fields.

        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from matplotlib import patches
        >>> from mpl_toolkits.axes_grid1 import make_axes_locatable
        >>> from geoana.em.static import MagnetostaticSphere

        Define the sphere.

        >>> mu_sphere = 10. ** -1
        >>> mu_background = 10. ** -3
        >>> radius = 1.0
        >>> simulation = MagnetostaticSphere(
        >>>     location=None, mu_sphere=mu_sphere, mu_background=mu_background, radius=radius, primary_field=None
        >>> )

        Now we create a set of gridded locations and compute the magnetic fields.

        >>> X, Y = np.meshgrid(np.linspace(-2*radius, 2*radius, 20), np.linspace(-2*radius, 2*radius, 20))
        >>> Z = np.zeros_like(X) + 0.25
        >>> xyz = np.stack((X, Y, Z), axis=-1)
        >>> ht = simulation.magnetic_field(xyz, field='total')
        >>> hs = simulation.magnetic_field(xyz, field='secondary')

        Finally, we plot the total and secondary magnetic fields.

        >>> fig, axs = plt.subplots(1, 2, figsize=(18,12))
        >>> titles = ['Total Magnetic Field', 'Secondary Magnetic Field']
        >>> for ax, H, title in zip(axs.flatten(), [ht, hs], titles):
        >>>     H_amp = np.linalg.norm(H, axis=-1)
        >>>     im = ax.pcolor(X, Y, H_amp, shading='auto')
        >>>     divider = make_axes_locatable(ax)
        >>>     cax = divider.append_axes("right", size="5%", pad=0.05)
        >>>     cb = plt.colorbar(im, cax=cax)
        >>>     cb.set_label(label= 'Amplitude ($A/m$)')
        >>>     ax.streamplot(X, Y, H[..., 0], H[..., 1], density=0.75)
        >>>     ax.add_patch(patches.Circle((0, 0), radius, fill=False, linestyle='--'))
        >>>     ax.set_ylabel('Y coordinate ($m$)')
        >>>     ax.set_xlabel('X coordinate ($m$)')
        >>>     ax.set_aspect('equal')
        >>>     ax.set_title(title)
        >>> plt.tight_layout()
        >>> plt.show()
        """
        xyz = check_xyz_dim(xyz)
        mu0 = self.mu_background
        mu1 = self.mu_sphere
        H0 = self.primary_field
        mu_cur = (mu1 - mu0) / (mu1 + 2 * mu0)
        r_vec = xyz - self.location
        r = np.linalg.norm(r_vec, axis=-1)

        Ht = np.zeros((*r.shape, 3))
        ind0 = r > self.radius

        # total field outside the sphere
        Ht[ind0] = H0 * (1. - mu_cur * self.radius ** 3. / r[ind0, None] ** 3) +\
            3. * (r_vec[ind0] @ H0)[..., None] * mu_cur * self.radius * r_vec[ind0] / r[ind0, None] ** 4

        # inside the sphere
        Ht[~ind0] = 3. * mu0 / (mu1 + 2. * mu0) * H0

        if field == 'total':
            return Ht

        if field != 'total':
            Hp = np.zeros((*r.shape, 3))
            Hp = Hp + H0
            if field == 'primary':
                return Hp

        # field was not primary or total
        Hs = Ht - Hp
        if field == 'secondary':
            return Hs

        return Ht, Hp, Hs

    def magnetic_flux_density(self, xyz, field='all'):
        """Magnetic flux density for a permeable sphere in a uniform magnetostatic field.

        .. math::

            \\mathbf{B} = \\mu \\mathbf{H}

        Parameters
        ----------
        xyz : (..., 3) numpy.ndarray
            Locations to evaluate at in units m.
        field : {'all', 'total', 'primary', 'secondary'}

        Returns
        -------
        Bt, Bp, Bs : (..., 3) np.ndarray
            If field == "all"
        B : (..., 3) np.ndarray
            If only requesting a single field.

        Examples
        --------
        Here, we define a sphere with permeability mu_sphere in a uniform magnetostatic field with permeability
        mu_background and plot the total and secondary magnetic flux densities.

        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from matplotlib import patches
        >>> from mpl_toolkits.axes_grid1 import make_axes_locatable
        >>> from geoana.em.static import MagnetostaticSphere

        Define the sphere.

        >>> mu_sphere = 10. ** -1
        >>> mu_background = 10. ** -3
        >>> radius = 1.0
        >>> simulation = MagnetostaticSphere(
        >>>     location=None, mu_sphere=mu_sphere, mu_background=mu_background, radius=radius, primary_field=None
        >>> )

        Now we create a set of gridded locations and compute the magnetic flux densities.

        >>> X, Y = np.meshgrid(np.linspace(-2*radius, 2*radius, 20), np.linspace(-2*radius, 2*radius, 20))
        >>> Z = np.zeros_like(X) + 0.25
        >>> xyz = np.stack((X, Y, Z), axis=-1)
        >>> bt = simulation.magnetic_flux_density(xyz, field='total')
        >>> bs = simulation.magnetic_flux_density(xyz, field='secondary')

        Finally, we plot the total and secondary magnetic flux densities.

        >>> fig, axs = plt.subplots(1, 2, figsize=(18,12))
        >>> titles = ['Total Magnetic Flux Density', 'Secondary Magnetic Flux Density']
        >>> for ax, B, title in zip(axs.flatten(), [bt, bs], titles):
        >>>     B_amp = np.linalg.norm(B, axis=-1)
        >>>     im = ax.pcolor(X, Y, B_amp, shading='auto')
        >>>     divider = make_axes_locatable(ax)
        >>>     cax = divider.append_axes("right", size="5%", pad=0.05)
        >>>     cb = plt.colorbar(im, cax=cax)
        >>>     cb.set_label(label= 'Amplitude (T)')
        >>>     ax.streamplot(X, Y, B[..., 0], B[..., 1], density=0.75)
        >>>     ax.add_patch(patches.Circle((0, 0), radius, fill=False, linestyle='--'))
        >>>     ax.set_ylabel('Y coordinate ($m$)')
        >>>     ax.set_xlabel('X coordinate ($m$)')
        >>>     ax.set_aspect('equal')
        >>>     ax.set_title(title)
        >>> plt.tight_layout()
        >>> plt.show()
        """

        xyz = check_xyz_dim(xyz)
        r = np.linalg.norm(xyz - self.location, axis=-1)

        mu = np.full(r.shape, self.mu_background)
        mu[r <= self.radius] = self.mu_sphere

        Ht = self.magnetic_field(xyz, field='total')
        Bt = mu[..., None] * Ht
        if field == 'total':
            return Bt

        if field != 'total':
            Hp = self.magnetic_field(xyz, field='primary')
            Bp = self.mu_background * Hp
            if field == 'primary':
                return Bp

        Bs = Bt - Bp
        if field == 'secondary':
            return Bs

        return Bt, Bp, Bs



from geoana.em.base import BaseEM
import numpy as np
from scipy.constants import mu_0


###############################################################################
#                                                                             #
#                           Utility Functions                                 #
#                                                                             #
###############################################################################


def peak_time(z, sigma, mu=mu_0):
    r"""Compute peak time for a plane wave in a homogeneous electromagnetic media.

    For a particular distance along the propagation path of a planewave in
    a homogeneous media, the `peak time <https://em.geosci.xyz/content/maxwell1_fundamentals/transient_planewaves_homogeneous/peaktime.html>`_
    is the time at which the maximum signal amplitude is observed. In other words,
    it is simple computation for when the peak of the planewave will arrive.

    For the quasistatic case (no electric displacement), the peak time is given by:

    .. math::
        t_{max} = \frac{\mu \sigma z^2}{6}

    where :math:`\mu` is the magnetic permeability, :math:`\sigma` is the electrical
    conductivity and *z* is the propagation distance.

    Notes
    -----
    The inputs values will be broadcasted together following normal numpy rules, and
    will support general shapes. Therefore every input, except for the `secondary` flag,
    can be arrays of the same shape.

    Parameters
    ----------
    z : float or numpy.ndarray
        propagation distance from the planewave source (m)
    sigma : float or numpy.ndarray
        electrical conductivity (S/m)
    mu : float or numpy.ndarray, optional
        magnetic permeability (A/m). Default is the permeability of free-space
        (:math:`\mu_0`)

    Returns
    -------
    float, np.ndarray
        Peak time/times in seconds. The dimensions for return will depend on
        whether floats or arrays were used to define the propagation distance
        and physical properties of the medium.

    """
    return (mu * sigma * z**2)/6.


def diffusion_distance(time, sigma, mu=mu_0):
    r"""Compute diffusion distance for a plane wave in a homogeneous electromagnetic media.

    For a planewave source in a homogeneous media, the `diffusion distance <https://em.geosci.xyz/content/maxwell1_fundamentals/transient_planewaves_homogeneous/peakdistance.html>`_
    is the propagation distance the peak has travelled at a given time. It is sometimes
    referred to as the peak distance. The diffusion distance is a simple computation for
    how far a planewave has travelled after a certain time.

    For the quasistatic case (no electric displacement), the diffusion distance is given by:

    .. math::
        D = \sqrt{\frac{2 t}{\mu \sigma}}

    where :math:`\mu` is the magnetic permeability, :math:`\sigma` is the electrical
    conductivity and *t* is the time.

    Notes
    -----
    The inputs values will be broadcasted together following normal numpy rules, and
    will support general shapes. Therefore every input, except for the `secondary` flag,
    can be arrays of the same shape.

    Parameters
    ----------
    time : float or numpy.ndarray
        propagation time (s)
    sigma : float or numpy.ndarray
        electrical conductivity (S/m)
    mu : float or numpy.ndarray, optional
        magnetic permeability (A/m). Default is the permeability of free-space
        (:math:`\mu_0`)

    Returns
    -------
    float, np.ndarray
        Diffusion distance in meters. The dimensions for return will depend on
        whether floats or arrays were used to define the propagation distance
        and physical properties of the medium.

    """
    return np.sqrt(2*time/(mu*sigma))


def theta(time, sigma, mu=mu_0):
    r"""
    Analog to wavenumber in the frequency domain. See Ward and Hohmann, 1988
    pages 174-175:

    .. math::
        \theta = \sqrt{\frac{\mu \sigma}{4t}}

    Parameters
    ----------
    time : float or numpy.ndarray
        propagation time (s)
    sigma : float or numpy.ndarray
        electrical conductivity (S/m)
    mu : float or numpy.ndarray, optional
        magnetic permeability (A/m). Default is the permeability of free-space
        (:math:`\mu_0`)

    Returns
    -------
    float, np.ndarray
        The theta value for all times.

    """
    return np.sqrt(mu*sigma/(4.*time))


class BaseTDEM(BaseEM):
    r"""
    Base time domain electromagnetic class

    The base TDEM class is contructed to define rudimentary properties and methods
    for time-domain electromagnetics.

    Parameters
    ----------
    time: int, float or (n_time) numpy.ndarray
        Time or times used for all computations in seconds.


    """

    def __init__(self, time, **kwargs):

        self.time = time
        super().__init__(**kwargs)

    @property
    def time(self):
        """Time (s) used for all computations

        Returns
        -------
        numpy.ndarray
            Time (or times) in seconds used for all computations
        """
        return self._time

    @time.setter
    def time(self, value):

        # Ensure float or numpy array of float
        try:
            value = np.atleast_1d(value).astype(float)
        except:
            raise TypeError(f"times are not a valid type")

        # Enforce positivity and dimensions
        if (value < 0.).any():
            raise ValueError("All times must be greater than 0")
        if value.ndim > 1:
            raise TypeError(f"times must be ('*') array")

        self._time = value

    def peak_time(self, z):
        r"""Compute peak time for a plane wave in a homogeneous electromagnetic media.

        For a particular distance along the propagation path of a planewave in
        a homogeneous media, the `peak time <https://em.geosci.xyz/content/maxwell1_fundamentals/transient_planewaves_homogeneous/peaktime.html>`_
        is the time at which the maximum signal amplitude is observed. In other words,
        it is simple computation for when the peak of the planewave will arrive.

        For the quasistatic case (no electric displacement), the peak time is given by:

        .. math::
            t_{max} = \frac{\mu \sigma z^2}{6}

        where :math:`\mu` is the magnetic permeability, :math:`\sigma` is the electrical
        conductivity and *z* is the propagation distance.

        Parameters
        ----------
        z : float or numpy.ndarray
            propagation distance from the planewave source (m)

        Returns
        -------
        float, np.ndarray
            Peak time/times in seconds

        """
        return peak_time(z, self.sigma, self.mu)

    @property
    def diffusion_distance(self):
        r"""Compute diffusion distance for a plane wave in a homogeneous electromagnetic media.

        For a planewave source in a homogeneous media, the `diffusion distance <https://em.geosci.xyz/content/maxwell1_fundamentals/transient_planewaves_homogeneous/peakdistance.html>`_
        is the propagation distance the peak has travelled at a given time. It is sometimes
        referred to as the peak distance. The diffusion distance is a simple computation for
        how far a planewave has travelled after a certain time.

        For the quasistatic case (no electric displacement), the diffusion distance is given by:

        .. math::
            D = \sqrt{\frac{2 t}{\mu \sigma}}

        where :math:`\mu` is the magnetic permeability, :math:`\sigma` is the electrical
        conductivity and *t* is the time.

        Returns
        -------
        float, np.ndarray
            Diffusion distance at all times.

        """
        return diffusion_distance(self.time, self.sigma, self.mu)

    @property
    def theta(self):
        r"""
        Analog to wavenumber in the frequency domain. See Ward and Hohmann, 1988
        pages 174-175:

        .. math::
            \theta = \sqrt{\frac{\mu \sigma}{4t}}

        Parameters
        ----------
        time : float or numpy.ndarray
            propagation time (s)
        sigma : float or numpy.ndarray
            electrical conductivity (S/m)
        mu : float or numpy.ndarray, optional
            magnetic permeability (A/m). Default is the permeability of free-space
            (:math:`\mu_0`)

        Returns
        -------
        float, np.ndarray
            The theta value for all times.

        """
        return theta(self.time, self.sigma, self.mu)



import numpy as np
from scipy.constants import mu_0

from ..base import BaseLineCurrent


class LineCurrentFreeSpace(BaseLineCurrent):
    """Class for a static line current in free space.

    The ``LineCurrentFreeSpace`` class is used to analytically compute the
    fields and potentials within freespace due to a set of static current-carrying wires.
    """

    def magnetic_field(self, xyz):
        r"""Compute the magnetic field for the static current-carrying wire segments.

        Parameters
        ----------
        xyz : (n, 3) numpy.ndarray xyz
            gridded locations at which we are calculating the magnetic field

        Returns
        -------
        (n, 3) numpy.ndarray
            The magnetic field at each observation location in H/m.

        Examples
        --------
        Here, we define a horizontal square loop and plot the magnetic field
        on the xz-plane that intercepts at y=0.

        >>> from geoana.em.static import LineCurrentFreeSpace
        >>> from geoana.utils import ndgrid
        >>> from geoana.plotting_utils import plot2Ddata
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt

        Let us begin by defining the loop. Note that to create an inductive
        source, we closed the loop.

        >>> x_nodes = np.array([-0.5, 0.5, 0.5, -0.5, -0.5])
        >>> y_nodes = np.array([-0.5, -0.5, 0.5, 0.5, -0.5])
        >>> z_nodes = np.zeros_like(x_nodes)
        >>> nodes = np.c_[x_nodes, y_nodes, z_nodes]
        >>> simulation = LineCurrentFreeSpace(nodes)

        Now we create a set of gridded locations and compute the magnetic field.

        >>> xyz = ndgrid(np.linspace(-1, 1, 50), np.array([0]), np.linspace(-1, 1, 50))
        >>> H = simulation.magnetic_field(xyz)

        Finally, we plot the magnetic field.

        >>> fig = plt.figure(figsize=(4, 4))
        >>> ax = fig.add_axes([0.15, 0.15, 0.75, 0.75])
        >>> plot2Ddata(xyz[:, [0, 2]], H[:, [0, 2]], ax=ax, vec=True, scale='log', ncontour=25)
        >>> ax.set_xlabel('X')
        >>> ax.set_ylabel('Z')
        >>> ax.set_title('Magnetic field')

        """

        # TRANSMITTER NODES
        I = self.current
        tx_nodes = self.nodes
        x1tr = tx_nodes[:, 0]
        x2tr = tx_nodes[:, 1]
        x3tr = tx_nodes[:, 2]

        n_loc = np.shape(xyz)[0]
        n_segments = self.n_segments

        hx0 = np.zeros(n_loc)
        hy0 = np.zeros(n_loc)
        hz0 = np.zeros(n_loc)

        for pp in range(0, n_segments):

            # Wire ends for transmitter wire pp
            x1a = x1tr[pp]
            x2a = x2tr[pp]
            x3a = x3tr[pp]
            x1b = x1tr[pp + 1]
            x2b = x2tr[pp + 1]
            x3b = x3tr[pp + 1]

            # Vector Lengths between points
            vab = np.sqrt((x1b - x1a) ** 2 + (x2b - x2a) ** 2 + (x3b - x3a) ** 2)
            vap = np.sqrt(
                (xyz[:, 0] - x1a) ** 2 + (xyz[:, 1] - x2a) ** 2 + (xyz[:, 2] - x3a) ** 2
            )
            vbp = np.sqrt(
                (xyz[:, 0] - x1b) ** 2 + (xyz[:, 1] - x2b) ** 2 + (xyz[:, 2] - x3b) ** 2
            )

            # Cosines from cos()=<v1,v2>/(|v1||v2|)
            cos_alpha = (
                (xyz[:, 0] - x1a) * (x1b - x1a)
                + (xyz[:, 1] - x2a) * (x2b - x2a)
                + (xyz[:, 2] - x3a) * (x3b - x3a)
            ) / (vap * vab)
            cos_beta = (
                (xyz[:, 0] - x1b) * (x1a - x1b)
                + (xyz[:, 1] - x2b) * (x2a - x2b)
                + (xyz[:, 2] - x3b) * (x3a - x3b)
            ) / (vbp * vab)

            # Determining Radial Vector From Wire
            dot_temp = (
                (x1a - xyz[:, 0]) * (x1b - x1a)
                + (x2a - xyz[:, 1]) * (x2b - x2a)
                + (x3a - xyz[:, 2]) * (x3b - x3a)
            )

            rx1 = (x1a - xyz[:, 0]) - dot_temp * (x1b - x1a) / vab ** 2
            rx2 = (x2a - xyz[:, 1]) - dot_temp * (x2b - x2a) / vab ** 2
            rx3 = (x3a - xyz[:, 2]) - dot_temp * (x3b - x3a) / vab ** 2

            r = np.sqrt(rx1 ** 2 + rx2 ** 2 + rx3 ** 2)

            phi = (cos_alpha + cos_beta) / r

            # I/4*pi in each direction
            ix1 = I * (x1b - x1a) / (4 * np.pi * vab)
            ix2 = I * (x2b - x2a) / (4 * np.pi * vab)
            ix3 = I * (x3b - x3a) / (4 * np.pi * vab)

            # Add contribution from wire pp into array
            hx0 = hx0 + phi * (-ix2 * rx3 + ix3 * rx2) / r
            hy0 = hy0 + phi * (ix1 * rx3 - ix3 * rx1) / r
            hz0 = hz0 + phi * (-ix1 * rx2 + ix2 * rx1) / r

        return np.c_[hx0, hy0, hz0]

    def magnetic_flux_density(self, xyz):
        r"""Compute the magnetic flux density for the static current-carrying wire segments.

        Parameters
        ----------
        xyz : (n, 3) numpy.ndarray xyz
            gridded locations at which we are calculating the magnetic flux density

        Returns
        -------
        (n, 3) numpy.ndarray
            The magnetic flux density at each observation location in T.

        Examples
        --------
        Here, we define a horizontal square loop and plot the magnetic flux
        density on the XZ-plane that intercepts at Y=0.

        >>> from geoana.em.static import LineCurrentFreeSpace
        >>> from geoana.utils import ndgrid
        >>> from geoana.plotting_utils import plot2Ddata
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt

        Let us begin by defining the loop. Note that to create an inductive
        source, we closed the loop

        >>> x_nodes = np.array([-0.5, 0.5, 0.5, -0.5, -0.5])
        >>> y_nodes = np.array([-0.5, -0.5, 0.5, 0.5, -0.5])
        >>> z_nodes = np.zeros_like(x_nodes)
        >>> nodes = np.c_[x_nodes, y_nodes, z_nodes]
        >>> simulation = LineCurrentFreeSpace(nodes)

        Now we create a set of gridded locations and compute the magnetic flux density.

        >>> xyz = ndgrid(np.linspace(-1, 1, 50), np.array([0]), np.linspace(-1, 1, 50))
        >>> B = simulation.magnetic_flux_density(xyz)

        Finally, we plot the magnetic flux density on the plane.

        >>> fig = plt.figure(figsize=(4, 4))
        >>> ax = fig.add_axes([0.15, 0.15, 0.8, 0.8])
        >>> plot2Ddata(xyz[:, [0, 2]], B[:, [0, 2]], ax=ax, vec=True, scale='log', ncontour=25)
        >>> ax.set_xlabel('X')
        >>> ax.set_ylabel('Z')
        >>> ax.set_title('Magnetic flux density')

        """

        return mu_0 * self.magnetic_field(xyz)



import numpy as np
from geoana.em.fdem.base import BaseFDEM
from geoana.spatial import repeat_scalar
from geoana.em.base import BaseElectricDipole, BaseMagneticDipole


class ElectricDipoleWholeSpace(BaseFDEM, BaseElectricDipole):
    r"""Class for simulating the fields and fluxes for a harmonic electric dipole in a wholespace.

    Harmonic electric dipole in a whole space. The source is
    (c.f. Ward and Hohmann, 1988 page 173). The source current
    density for a dipole located at :math:`\mathbf{r}_s` with orientation
    :math:`\mathbf{\hat{u}}`

    .. math::
        \mathbf{J}(\mathbf{r}) = I ds \delta(\mathbf{r} - \mathbf{r}_s)\mathbf{\hat{u}}

    """
    def vector_potential(self, xyz):
        r"""Vector potential for the harmonic current dipole at a set of gridded locations.

        For an electric current dipole oriented in the :math:`\hat{u}` direction with
        dipole moment :math:`I ds`, the magnetic vector potential at frequency :math:`f`
        at vector distance :math:`\mathbf{r}` from the dipole is given by:

        .. math::
            \mathbf{a}(\mathbf{r}) = \frac{I ds}{4 \pi r} e^{-ikr} \mathbf{\hat{u}}

        where

        .. math::
            k = \sqrt{\omega^2 \mu \varepsilon - i \omega \mu \sigma}

        Parameters
        ----------
        xyz : (..., 3) numpy.ndarray
            Gridded xyz locations

        Returns
        -------
        (n_freq, ..., 3) numpy.array of complex
            Magnetic vector potential at all frequencies for the gridded
            locations provided. Output array is squeezed when n_freq and/or
            n_loc = 1.

        Examples
        --------
        Here, we define an z-oriented electric dipole and plot the magnetic
        vector potential on the xy-plane that intercepts z=0.

        >>> from geoana.em.fdem import ElectricDipoleWholeSpace
        >>> from geoana.utils import ndgrid
        >>> from geoana.plotting_utils import plot2Ddata
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt

        Let us begin by defining the electric current dipole.

        >>> frequency = np.logspace(1, 3, 3)
        >>> location = np.r_[0., 0., 0.]
        >>> orientation = np.r_[0., 0., 1.]
        >>> current = 1.
        >>> sigma = 1.0
        >>> simulation = ElectricDipoleWholeSpace(
        >>>     frequency, location=location, orientation=orientation,
        >>>     current=current, sigma=sigma
        >>> )

        Now we create a set of gridded locations and compute the vector potential.

        >>> xyz = ndgrid(np.linspace(-1, 1, 20), np.linspace(-1, 1, 20), np.array([0]))
        >>> a = simulation.vector_potential(xyz)

        Finally, we plot the real and imaginary components of the vector potential.
        Given the symmetry, there are only vertical components.

        >>> f_ind = 1
        >>> fig = plt.figure(figsize=(6, 3))
        >>> ax1 = fig.add_axes([0.15, 0.15, 0.40, 0.75])
        >>> plot2Ddata(xyz[:, 0:2], np.real(a[f_ind, :, 2]), ax=ax1, scale='log', ncontour=25)
        >>> ax1.set_xlabel('X')
        >>> ax1.set_ylabel('Z')
        >>> ax1.autoscale(tight=True)
        >>> ax1.set_title('Real component {} Hz'.format(frequency[f_ind]))
        >>> ax2 = fig.add_axes([0.6, 0.15, 0.40, 0.75])
        >>> plot2Ddata(xyz[:, 0:2], np.imag(a[f_ind, :, 2]), ax=ax2, scale='log', ncontour=25)
        >>> ax2.set_xlabel('X')
        >>> ax2.set_yticks([])
        >>> ax2.set_title('Imag component {} Hz'.format(frequency[f_ind]))
        >>> ax2.autoscale(tight=True)

        """
        r = np.linalg.norm(xyz - self.location, axis=-1)
        k = self.wavenumber
        for dim in range(r.ndim):
            k = k[:, None]

        a = self.current * self.length / (4*np.pi*r) * np.exp(-1j*k * r)
        return (a[..., None] * self.orientation).squeeze()

    def electric_field(self, xyz):
        r"""Electric field for the harmonic current dipole at a set of gridded locations.

        For an electric current dipole oriented in the :math:`\hat{u}` direction with
        dipole moment :math:`I ds` and harmonic frequency :math:`f`, this method computes the
        electric field at the set of gridded xyz locations provided.

        The analytic solution is adapted from Ward and Hohmann (1988). For a harmonic electric
        current dipole oriented in the :math:`\hat{x}` direction, the solution at vector distance
        :math:`\mathbf{r}` from the current dipole is:

        .. math::
            \;\;\;\;\;\; \mathbf{E}(\mathbf{r}) = \frac{I ds}{4\pi (\sigma + i \omega \varepsilon) r^3} & e^{-ikr} ... \\
            & \Bigg [ \Bigg ( \frac{x^2}{r^2}\hat{x} + \frac{xy}{r^2}\hat{y} + \frac{xz}{r^2}\hat{z} \Bigg ) \big ( -k^2r^2 + 3ikr + 3 \big ) + \big ( k^2 r^2 - ikr - 1 \big ) \hat{x} \Bigg ]

        where

        .. math::
            k = \sqrt{\omega^2 \mu \varepsilon - i \omega \mu \sigma}

        Parameters
        ----------
        xyz : (..., 3) numpy.ndarray
            Gridded xyz locations

        Returns
        -------
        (n_freq, ..., 3) numpy.array of complex
            Electric field at all frequencies for the gridded
            locations provided. Output array is squeezed when n_freq and/or
            n_loc = 1.

        Examples
        --------
        Here, we define an x-oriented electric dipole and plot the electric
        field on the xz-plane that intercepts y=0.

        >>> from geoana.em.fdem import ElectricDipoleWholeSpace
        >>> from geoana.utils import ndgrid
        >>> from geoana.plotting_utils import plot2Ddata
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt

        Let us begin by defining the electric current dipole.

        >>> frequency = np.logspace(1, 3, 3)
        >>> location = np.r_[0., 0., 0.]
        >>> orientation = np.r_[1., 0., 0.]
        >>> current = 1.
        >>> sigma = 1.0
        >>> simulation = ElectricDipoleWholeSpace(
        >>>     frequency, location=location, orientation=orientation,
        >>>     current=current, sigma=sigma
        >>> )

        Now we create a set of gridded locations and compute the electric field.

        >>> xyz = ndgrid(np.linspace(-1, 1, 20), np.array([0]), np.linspace(-1, 1, 20))
        >>> E = simulation.electric_field(xyz)

        Finally, we plot the real and imaginary components of the electric field

        >>> f_ind = 1
        >>> fig = plt.figure(figsize=(6, 3))
        >>> ax1 = fig.add_axes([0.15, 0.15, 0.40, 0.75])
        >>> plot2Ddata(
        >>>     xyz[:, 0::2], np.real(E[f_ind, :, 0::2]), vec=True, ax=ax1, scale='log', ncontour=25
        >>> )
        >>> ax1.set_xlabel('X')
        >>> ax1.set_ylabel('Z')
        >>> ax1.autoscale(tight=True)
        >>> ax1.set_title('Real component {} Hz'.format(frequency[f_ind]))
        >>> ax2 = fig.add_axes([0.6, 0.15, 0.40, 0.75])
        >>> plot2Ddata(
        >>>     xyz[:, 0::2], np.imag(E[f_ind, :, 0::2]), vec=True, ax=ax2, scale='log', ncontour=25
        >>> )
        >>> ax2.set_xlabel('X')
        >>> ax2.set_yticks([])
        >>> ax2.autoscale(tight=True)
        >>> ax2.set_title('Imag component {} Hz'.format(frequency[f_ind]))

        """
        dxyz = xyz - self.location
        r = np.linalg.norm(dxyz, axis=-1)
        k = self.wavenumber
        for dim in range(r.ndim):
            k = k[:, None]

        kr = k * r
        ikr = 1j * kr

        front_term = (
            (self.current * self.length) / (4 * np.pi * self.sigma * r**3) *
            np.exp(-ikr)
        )
        symmetric_term = (
            ((dxyz @ self.orientation) * (-kr**2 + 3*ikr + 3) / r**2)[..., None] * dxyz
        )
        oriented_term = (kr**2 - ikr - 1)[..., None] * self.orientation

        return (front_term[..., None] * (symmetric_term + oriented_term)).squeeze()

    def current_density(self, xyz):
        r"""Current density for the harmonic current dipole at a set of gridded locations.

        For an electric current dipole oriented in the :math:`\hat{u}` direction with
        dipole moment :math:`I ds` and harmonic frequency :math:`f`, this method computes the
        current density at the set of gridded xyz locations provided.

        The analytic solution is adapted from Ward and Hohmann (1988). For a harmonic electric
        current dipole oriented in the :math:`\hat{x}` direction, the solution at vector distance
        :math:`\mathbf{r}` from the current dipole is:

        .. math::
            \;\;\;\; \mathbf{J}(\mathbf{r}) = \frac{\sigma I ds}{4\pi (\sigma + i \omega \varepsilon) r^3} & e^{-ikr} ... \\
            & \Bigg [ \Bigg ( \frac{x^2}{r^2}\hat{x} + \frac{xy}{r^2}\hat{y} + \frac{xz}{r^2}\hat{z} \Bigg ) \big ( -k^2r^2 + 3ikr + 3 \big ) \big ( k^2 r^2 - ikr - 1 \big ) \hat{x} \Bigg ]

        where

        .. math::
            k = \sqrt{\omega^2 \mu \varepsilon - i \omega \mu \sigma}

        Parameters
        ----------
        xyz : (n, 3) numpy.ndarray
            Gridded xyz locations

        Returns
        -------
        (n_freq, n_loc, 3) numpy.array of complex
            Current density at all frequencies for the gridded
            locations provided. Output array is squeezed when n_freq and/or
            n_loc = 1.

        Examples
        --------
        Here, we define an x-oriented electric dipole and plot the current
        density on the xz-plane that intercepts y=0.

        >>> from geoana.em.fdem import ElectricDipoleWholeSpace
        >>> from geoana.utils import ndgrid
        >>> from geoana.plotting_utils import plot2Ddata
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt

        Let us begin by defining the electric current dipole.

        >>> frequency = np.logspace(1, 3, 3)
        >>> location = np.r_[0., 0., 0.]
        >>> orientation = np.r_[1., 0., 0.]
        >>> current = 1.
        >>> sigma = 1.0
        >>> simulation = ElectricDipoleWholeSpace(
        >>>     frequency, location=location, orientation=orientation,
        >>>     current=current, sigma=sigma
        >>> )

        Now we create a set of gridded locations and compute the current density.

        >>> xyz = ndgrid(np.linspace(-1, 1, 20), np.array([0]), np.linspace(-1, 1, 20))
        >>> J = simulation.current_density(xyz)

        Finally, we plot the real and imaginary components of the current density.

        >>> f_ind = 1
        >>> fig = plt.figure(figsize=(6, 3))
        >>> ax1 = fig.add_axes([0.15, 0.15, 0.40, 0.75])
        >>> plot2Ddata(
        >>>     xyz[:, 0::2], np.real(J[f_ind, :, 0::2]), vec=True, ax=ax1, scale='log', ncontour=25
        >>> )
        >>> ax1.set_xlabel('X')
        >>> ax1.set_ylabel('Z')
        >>> ax1.autoscale(tight=True)
        >>> ax1.set_title('Real component {} Hz'.format(frequency[f_ind]))
        >>> ax2 = fig.add_axes([0.6, 0.15, 0.40, 0.75])
        >>> plot2Ddata(
        >>>     xyz[:, 0::2], np.imag(J[f_ind, :, 0::2]), vec=True, ax=ax2, scale='log', ncontour=25
        >>> )
        >>> ax2.set_xlabel('X')
        >>> ax2.set_yticks([])
        >>> ax2.autoscale(tight=True)
        >>> ax2.set_title('Imag component {} Hz'.format(frequency[f_ind]))

        """
        return self.sigma * self.electric_field(xyz)

    def magnetic_field(self, xyz):
        r"""Magnetic field produced by the harmonic current dipole at a set of gridded locations.

        For an electric current dipole oriented in the :math:`\hat{u}` direction with
        dipole moment :math:`I ds` and harmonic frequency :math:`f`, this method computes the
        magnetic field at the set of gridded xyz locations provided.

        The analytic solution is adapted from Ward and Hohmann (1988). For a harmonic electric
        current dipole oriented in the :math:`\hat{x}` direction, the solution at vector distance
        :math:`\mathbf{r}` from the current dipole is:

        .. math::
            \mathbf{H}(\mathbf{r}) = \frac{I ds}{4\pi r^2} (ikr + 1) e^{-ikr} \big ( - \frac{z}{r}\hat{y} + \frac{y}{r}\hat{z} \big )

        where

        .. math::
            k = \sqrt{\omega^2 \mu \varepsilon - i \omega \mu \sigma}

        Parameters
        ----------
        xyz : (..., 3) numpy.ndarray
            Gridded xyz locations

        Returns
        -------
        (n_freq, ..., 3) numpy.array of complex
            Magnetic field at all frequencies for the gridded
            locations provided. Output array is squeezed when n_freq and/or
            n_loc = 1.

        Examples
        --------
        Here, we define an z-oriented electric dipole and plot the magnetic field
        on the xy-plane that intercepts z=0.

        >>> from geoana.em.fdem import ElectricDipoleWholeSpace
        >>> from geoana.utils import ndgrid
        >>> from geoana.plotting_utils import plot2Ddata
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt

        Let us begin by defining the electric current dipole.

        >>> frequency = np.logspace(1, 3, 3)
        >>> location = np.r_[0., 0., 0.]
        >>> orientation = np.r_[0., 0., 1.]
        >>> current = 1.
        >>> sigma = 1.0
        >>> simulation = ElectricDipoleWholeSpace(
        >>>     frequency, location=location, orientation=orientation,
        >>>     current=current, sigma=sigma
        >>> )

        Now we create a set of gridded locations and compute the magnetic field.

        >>> xyz = ndgrid(np.linspace(-1, 1, 20), np.linspace(-1, 1, 20), np.array([0]))
        >>> H = simulation.magnetic_field(xyz)

        Finally, we plot the real and imaginary components of the magnetic field.

        >>> f_ind = 1
        >>> fig = plt.figure(figsize=(6, 3))
        >>> ax1 = fig.add_axes([0.15, 0.15, 0.40, 0.75])
        >>> plot2Ddata(
        >>>     xyz[:, 0:2], np.real(H[f_ind, :, 0:2]), vec=True, ax=ax1, scale='log', ncontour=25
        >>> )
        >>> ax1.set_xlabel('X')
        >>> ax1.set_ylabel('Y')
        >>> ax1.autoscale(tight=True)
        >>> ax1.set_title('Real component {} Hz'.format(frequency[f_ind]))
        >>> ax2 = fig.add_axes([0.6, 0.15, 0.40, 0.75])
        >>> plot2Ddata(
        >>>     xyz[:, 0:2], np.imag(H[f_ind, :, 0:2]), vec=True, ax=ax2, scale='log', ncontour=25
        >>> )
        >>> ax2.set_xlabel('X')
        >>> ax2.set_yticks([])
        >>> ax2.autoscale(tight=True)
        >>> ax2.set_title('Imag component {} Hz'.format(frequency[f_ind]))

        """
        dxyz = xyz - self.location
        r = np.linalg.norm(dxyz, axis=-1)
        k = self.wavenumber
        for dim in range(r.ndim):
            k = k[:, None]
        kr = k * r
        ikr = 1j*kr

        front_term = (
            self.current * self.length / (4 * np.pi * r**2) * (ikr + 1) *
            np.exp(-ikr)
        )
        return (-front_term[..., None] * np.cross(dxyz, self.orientation) / r[..., None]).squeeze()

    def magnetic_flux_density(self, xyz):
        r"""Magnetic flux density produced by the harmonic electric current dipole at a set of gridded locations.

        For an electric current dipole oriented in the :math:`\hat{u}` direction with
        dipole moment :math:`I ds` and harmonic frequency :math:`f`, this method computes the
        magnetic flux density at the set of gridded xyz locations provided.

        The analytic solution is adapted from Ward and Hohmann (1988). For a harmonic electric
        current dipole oriented in the :math:`\hat{x}` direction, the solution at vector distance
        :math:`\mathbf{r}` from the current dipole is:

        .. math::
            \mathbf{B}(\mathbf{r}) = \frac{\mu I ds}{4\pi r^2} (ikr + 1) e^{-ikr} \bigg ( - \frac{z}{r}\hat{y} + \frac{y}{r}\hat{z} \bigg )

        where

        .. math::
            k = \sqrt{\omega^2 \mu \varepsilon - i \omega \mu \sigma}

        Parameters
        ----------
        xyz : (..., 3) numpy.ndarray
            Gridded xyz locations

        Returns
        -------
        (n_freq, ..., 3) numpy.array of complex
            Magnetic flux at all frequencies for the gridded
            locations provided. Output array is squeezed when n_freq and/or
            n_loc = 1.

        Examples
        --------
        Here, we define an z-oriented electric dipole and plot the magnetic flux density
        on the xy-plane that intercepts z=0.

        >>> from geoana.em.fdem import ElectricDipoleWholeSpace
        >>> from geoana.utils import ndgrid
        >>> from geoana.plotting_utils import plot2Ddata
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt

        Let us begin by defining the electric current dipole.

        >>> frequency = np.logspace(1, 3, 3)
        >>> location = np.r_[0., 0., 0.]
        >>> orientation = np.r_[0., 0., 1.]
        >>> current = 1.
        >>> sigma = 1.0
        >>> simulation = ElectricDipoleWholeSpace(
        >>>     frequency, location=location, orientation=orientation,
        >>>     current=current, sigma=sigma
        >>> )

        Now we create a set of gridded locations and compute the magnetic flux density.

        >>> xyz = ndgrid(np.linspace(-1, 1, 20), np.linspace(-1, 1, 20), np.array([0]))
        >>> B = simulation.magnetic_flux_density(xyz)

        Finally, we plot the real and imaginary components of the magnetic flux density.

        >>> f_ind = 1
        >>> fig = plt.figure(figsize=(6, 3))
        >>> ax1 = fig.add_axes([0.15, 0.15, 0.40, 0.75])
        >>> plot2Ddata(
        >>>     xyz[:, 0:2], np.real(B[f_ind, :, 0:2]), vec=True, ax=ax1, scale='log', ncontour=25
        >>> )
        >>> ax1.set_xlabel('X')
        >>> ax1.set_ylabel('Y')
        >>> ax1.autoscale(tight=True)
        >>> ax1.set_title('Real component {} Hz'.format(frequency[f_ind]))
        >>> ax2 = fig.add_axes([0.6, 0.15, 0.40, 0.75])
        >>> plot2Ddata(
        >>>     xyz[:, 0:2], np.imag(B[f_ind, :, 0:2]), vec=True, ax=ax2, scale='log', ncontour=25
        >>> )
        >>> ax2.set_xlabel('X')
        >>> ax2.set_yticks([])
        >>> ax2.autoscale(tight=True)
        >>> ax2.set_title('Imag component {} Hz'.format(frequency[f_ind]))

        """
        return self.mu * self.magnetic_field(xyz)


class MagneticDipoleWholeSpace(BaseFDEM, BaseMagneticDipole):
    """
    Harmonic magnetic dipole in a whole space.
    """

    def vector_potential(self, xyz):
        r"""Vector potential for the harmonic magnetic dipole at a set of gridded locations.

        For a harmonic magnetic dipole oriented in the :math:`\hat{u}` direction with
        moment amplitude :math:`m`, the magnetic vector potential at frequency :math:`f`
        at vector distance :math:`\mathbf{r}` from the dipole is given by:

        .. math::
            \mathbf{a}(\mathbf{r}) = \frac{i \omega \mu m}{4 \pi r} e^{-ikr}
            \mathbf{\hat{u}}

        where

        .. math::
            k = \sqrt{\omega^2 \mu \varepsilon - i \omega \mu \sigma}

        Parameters
        ----------
        xyz : (n, 3) numpy.ndarray
            Gridded xyz locations

        Returns
        -------
        (n_freq, n_loc, 3) numpy.array of complex
            Magnetic vector potential at all frequencies for the gridded
            locations provided. Output array is squeezed when n_freq and/or
            n_loc = 1.

        Examples
        --------
        Here, we define an z-oriented magnetic dipole and plot the magnetic
        vector potential on the xy-plane that intercepts z=0.

        >>> from geoana.em.fdem import MagneticDipoleWholeSpace
        >>> from geoana.utils import ndgrid
        >>> from geoana.plotting_utils import plot2Ddata
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt

        Let us begin by defining the electric current dipole.

        >>> frequency = np.logspace(1, 3, 3)
        >>> location = np.r_[0., 0., 0.]
        >>> orientation = np.r_[0., 0., 1.]
        >>> moment = 1.
        >>> sigma = 1.0
        >>> simulation = MagneticDipoleWholeSpace(
        >>>     frequency, location=location, orientation=orientation,
        >>>     moment=moment, sigma=sigma
        >>> )

        Now we create a set of gridded locations and compute the vector potential.

        >>> xyz = ndgrid(np.linspace(-1, 1, 20), np.linspace(-1, 1, 20), np.array([0]))
        >>> a = simulation.vector_potential(xyz)

        Finally, we plot the real and imaginary components of the vector potential.

        >>> f_ind = 1
        >>> fig = plt.figure(figsize=(6, 3))
        >>> ax1 = fig.add_axes([0.15, 0.15, 0.40, 0.75])
        >>> plot2Ddata(xyz[:, 0:2], np.real(a[f_ind, :, 2]), ax=ax1, scale='log', ncontour=25)
        >>> ax1.set_xlabel('X')
        >>> ax1.set_ylabel('Z')
        >>> ax1.autoscale(tight=True)
        >>> ax1.set_title('Real component {} Hz'.format(frequency[f_ind]))
        >>> ax2 = fig.add_axes([0.6, 0.15, 0.40, 0.75])
        >>> plot2Ddata(xyz[:, 0:2], np.imag(a[f_ind, :, 2]), ax=ax2, scale='log', ncontour=25)
        >>> ax2.set_xlabel('X')
        >>> ax2.set_yticks([])
        >>> ax2.autoscale(tight=True)
        >>> ax2.set_title('Imag component {} Hz'.format(frequency[f_ind]))

        """
        # r = self.distance(xyz)
        # f = (
        #     (1j * self.omega * self.mu * self.moment) / (4 * np.pi * r) *
        #     np.exp(-1j * self.wavenumber * r)
        # )
        # f = np.kron(np.ones(1, 3), np.atleast_2d(f).T)
        # return self.dot_orientation(f)

        n_freq = len(self.frequency)
        n_loc = np.shape(xyz)[0]

        r = self.distance(xyz)
        k = self.wavenumber

        tile_r = np.tile(r.reshape((1, n_loc)), (n_freq, 1))
        tile_w = np.tile(self.omega.reshape((n_freq, 1)), (1, n_loc))

        a = (1j * tile_w * self.mu * self.moment) * (
            1 / (4*np.pi*tile_r) * np.exp(-1j*np.outer(k, r))
        )

        v = self.orientation.reshape(1, 1, 3)
        a = a.reshape((n_freq, n_loc, 1))

        return np.kron(v, a).squeeze()

    def electric_field(self, xyz):
        r"""Electric field for the harmonic magnetic dipole at a set of gridded locations.

        For a harmonic magnetic dipole oriented in the :math:`\hat{u}` direction with
        moment amplitude :math:`m` and harmonic frequency :math:`f`, this method computes the
        electric field at the set of gridded xyz locations provided.

        The analytic solution is adapted from Ward and Hohmann (1988). For a harmonic electric
        magnetic dipole oriented in the :math:`\hat{x}` direction, the solution at vector distance
        :math:`\mathbf{r}` from the dipole is:

        .. math::
            \mathbf{E}(\mathbf{r}) = \frac{i\omega \mu m}{4\pi r^2} (ikr + 1) e^{-ikr} \big ( - \frac{z}{r}\hat{y} + \frac{y}{r}\hat{z} \big )

        where

        .. math::
            k = \sqrt{\omega^2 \mu \varepsilon - i \omega \mu \sigma}

        Parameters
        ----------
        xyz : (n, 3) numpy.ndarray
            Gridded xyz locations

        Returns
        -------
        (n_freq, n_loc, 3) numpy.array of complex
            Electric field at all frequencies for the gridded
            locations provided. Output array is squeezed when n_freq and/or
            n_loc = 1.

        Examples
        --------
        Here, we define an z-oriented magnetic dipole and plot the electric
        field on the xy-plane that intercepts z=0.

        >>> from geoana.em.fdem import MagneticDipoleWholeSpace
        >>> from geoana.utils import ndgrid
        >>> from geoana.plotting_utils import plot2Ddata
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt

        Let us begin by defining the electric current dipole.

        >>> frequency = np.logspace(1, 3, 3)
        >>> location = np.r_[0., 0., 0.]
        >>> orientation = np.r_[0., 0., 1.]
        >>> moment = 1.
        >>> sigma = 1.0
        >>> simulation = MagneticDipoleWholeSpace(
        >>>     frequency, location=location, orientation=orientation,
        >>>     moment=moment, sigma=sigma
        >>> )

        Now we create a set of gridded locations and compute the electric field.

        >>> xyz = ndgrid(np.linspace(-1, 1, 20), np.linspace(-1, 1, 20), np.array([0]))
        >>> E = simulation.electric_field(xyz)

        Finally, we plot the real and imaginary components of the electric field

        >>> f_ind = 1
        >>> fig = plt.figure(figsize=(6, 3))
        >>> ax1 = fig.add_axes([0.15, 0.15, 0.40, 0.75])
        >>> plot2Ddata(
        >>>     xyz[:, 0:2], np.real(E[f_ind, :, 0:2]), vec=True, ax=ax1, scale='log', ncontour=25
        >>> )
        >>> ax1.set_xlabel('X')
        >>> ax1.set_ylabel('Y')
        >>> ax1.autoscale(tight=True)
        >>> ax1.set_title('Real component {} Hz'.format(frequency[f_ind]))
        >>> ax2 = fig.add_axes([0.6, 0.15, 0.40, 0.75])
        >>> plot2Ddata(
        >>>     xyz[:, 0:2], np.imag(E[f_ind, :, 0:2]), vec=True, ax=ax2, scale='log', ncontour=25
        >>> )
        >>> ax2.set_xlabel('X')
        >>> ax2.set_yticks([])
        >>> ax2.autoscale(tight=True)
        >>> ax2.set_title('Imag component {} Hz'.format(frequency[f_ind]))

        """
        # dxyz = self.vector_distance(xyz)
        # r = repeat_scalar(self.distance(xyz))
        # kr = self.wavenumber*r
        # ikr = 1j * kr

        # front_term = (
        #     (1j * self.omega * self.mu * self.moment) / (4. * np.pi * r**2) *
        #     (ikr + 1) * np.exp(-ikr)
        # )
        # return front_term * self.cross_orientation(dxyz) / r

        n_freq = len(self.frequency)
        n_loc = np.shape(xyz)[0]

        k = self.wavenumber
        r = self.distance(xyz)

        # (n_freq, n_loc)
        tile_r = np.tile(r.reshape((1, n_loc)), (n_freq, 1))
        tile_w = np.tile(self.omega.reshape((n_freq, 1)), (1, n_loc))
        kr = np.outer(k, r)
        ikr = 1j * kr

        first_term = (
            (1j * tile_w * self.mu * self.moment) *
            (1 / (4 * np.pi * tile_r**2) * (ikr + 1) * np.exp(-ikr))
        ).reshape((n_freq, n_loc, 1))
        first_term = np.tile(first_term, (1, 1, 3))

        r = repeat_scalar(r)
        dxyz = self.vector_distance(xyz)

        second_term = (self.cross_orientation(dxyz) / r).reshape((1, n_loc, 3))
        second_term = np.tile(second_term, (n_freq, 1, 1))

        return (first_term * second_term).squeeze()


    def current_density(self, xyz):
        r"""Current density for the harmonic magnetic dipole at a set of gridded locations.

        For a harmonic magnetic dipole oriented in the :math:`\hat{u}` direction with
        moment amplitude :math:`m` and harmonic frequency :math:`f`, this method computes the
        current density at the set of gridded xyz locations provided.

        The analytic solution is adapted from Ward and Hohmann (1988). For a harmonic electric
        magnetic dipole oriented in the :math:`\hat{x}` direction, the solution at vector distance
        :math:`\mathbf{r}` from the dipole is:

        .. math::
            \mathbf{J}(\mathbf{r}) = \frac{i\omega \mu \sigma m}{4\pi r^2} (ikr + 1) e^{-ikr} \big ( - \frac{z}{r}\hat{y} + \frac{y}{r}\hat{z} \big )

        where

        .. math::
            k = \sqrt{\omega^2 \mu \varepsilon - i \omega \mu \sigma}

        Parameters
        ----------
        xyz : (n, 3) numpy.ndarray
            Gridded xyz locations

        Returns
        -------
        (n_freq, n_loc, 3) numpy.array of complex
            Current density at all frequencies for the gridded
            locations provided. Output array is squeezed when n_freq and/or
            n_loc = 1.

        Examples
        --------
        Here, we define an z-oriented magnetic dipole and plot the electric
        current density on the xy-plane that intercepts z=0.

        >>> from geoana.em.fdem import MagneticDipoleWholeSpace
        >>> from geoana.utils import ndgrid
        >>> from geoana.plotting_utils import plot2Ddata
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt

        Let us begin by defining the electric current dipole.

        >>> frequency = np.logspace(1, 3, 3)
        >>> location = np.r_[0., 0., 0.]
        >>> orientation = np.r_[0., 0., 1.]
        >>> moment = 1.
        >>> sigma = 1.0
        >>> simulation = MagneticDipoleWholeSpace(
        >>>     frequency, location=location, orientation=orientation,
        >>>     moment=moment, sigma=sigma
        >>> )

        Now we create a set of gridded locations and compute the current density.

        >>> xyz = ndgrid(np.linspace(-1, 1, 20), np.linspace(-1, 1, 20), np.array([0]))
        >>> J = simulation.current_density(xyz)

        Finally, we plot the real and imaginary components of the current density.

        >>> f_ind = 2
        >>> fig = plt.figure(figsize=(6, 3))
        >>> ax1 = fig.add_axes([0.15, 0.15, 0.40, 0.75])
        >>> plot2Ddata(
        >>>     xyz[:, 0:2], np.real(J[f_ind, :, 0:2]), vec=True, ax=ax1, scale='log', ncontour=25
        >>> )
        >>> ax1.set_xlabel('X')
        >>> ax1.set_ylabel('Y')
        >>> ax1.autoscale(tight=True)
        >>> ax1.set_title('Real component {} Hz'.format(frequency[f_ind]))
        >>> ax2 = fig.add_axes([0.6, 0.15, 0.40, 0.75])
        >>> plot2Ddata(
        >>>     xyz[:, 0:2], np.imag(J[f_ind, :, 0:2]), vec=True, ax=ax2, scale='log', ncontour=25
        >>> )
        >>> ax2.set_xlabel('X')
        >>> ax2.set_yticks([])
        >>> ax2.autoscale(tight=True)
        >>> ax2.set_title('Imag component {} Hz'.format(frequency[f_ind]))

        """
        return self.sigma * self.electric_field(xyz)

    def magnetic_field(self, xyz):
        r"""Magnetic field for the harmonic magnetic dipole at a set of gridded locations.

        For a harmonic magnetic dipole oriented in the :math:`\hat{u}` direction with
        moment amplitude :math:`m` and harmonic frequency :math:`f`, this method computes the
        magnetic field at the set of gridded xyz locations provided.

        The analytic solution is adapted from Ward and Hohmann (1988). For a harmonic magnetic
        dipole oriented in the :math:`\hat{x}` direction, the solution at vector distance
        :math:`\mathbf{r}` from the dipole is:

        .. math::
            \mathbf{H}(\mathbf{r}) = \frac{m}{4\pi r^3} e^{-ikr} \Bigg [ \Bigg ( \frac{x^2}{r^2}\hat{x} + \frac{xy}{r^2}\hat{y} + \frac{xz}{r^2}\hat{z} \Bigg ) ... \\
            \big ( -k^2r^2 + 3ikr + 3 \big ) \big ( k^2 r^2 - ikr - 1 \big ) \hat{x} \Bigg ]

        where

        .. math::
            k = \sqrt{\omega^2 \mu \varepsilon - i \omega \mu \sigma}

        Parameters
        ----------
        xyz : (n, 3) numpy.ndarray
            Gridded xyz locations

        Returns
        -------
        (n_freq, n_loc, 3) numpy.array of complex
            Magnetic field at all frequencies for the gridded
            locations provided. Output array is squeezed when n_freq and/or
            n_loc = 1.

        Examples
        --------
        Here, we define an z-oriented magnetic dipole and plot the magnetic field
        on the xz-plane that intercepts y=0.

        >>> from geoana.em.fdem import MagneticDipoleWholeSpace
        >>> from geoana.utils import ndgrid
        >>> from geoana.plotting_utils import plot2Ddata
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt

        Let us begin by defining the electric current dipole.

        >>> frequency = np.logspace(1, 3, 3)
        >>> location = np.r_[0., 0., 0.]
        >>> orientation = np.r_[0., 0., 1.]
        >>> moment = 1.
        >>> sigma = 1.0
        >>> simulation = MagneticDipoleWholeSpace(
        >>>     frequency, location=location, orientation=orientation,
        >>>     moment=moment, sigma=sigma
        >>> )

        Now we create a set of gridded locations and compute the magnetic field.

        >>> xyz = ndgrid(np.linspace(-1, 1, 20), np.array([0]), np.linspace(-1, 1, 20))
        >>> H = simulation.magnetic_field(xyz)

        Finally, we plot the real and imaginary components of the magnetic field.

        >>> f_ind = 2
        >>> fig = plt.figure(figsize=(6, 3))
        >>> ax1 = fig.add_axes([0.15, 0.15, 0.40, 0.75])
        >>> plot2Ddata(
        >>>     xyz[:, 0::2], np.real(H[f_ind, :, 0::2]), vec=True, ax=ax1, scale='log', ncontour=25
        >>> )
        >>> ax1.set_xlabel('X')
        >>> ax1.set_ylabel('Z')
        >>> ax1.autoscale(tight=True)
        >>> ax1.set_title('Real component {} Hz'.format(frequency[f_ind]))
        >>> ax2 = fig.add_axes([0.6, 0.15, 0.40, 0.75])
        >>> plot2Ddata(
        >>>     xyz[:, 0::2], np.imag(H[f_ind, :, 0::2]), vec=True, ax=ax2, scale='log', ncontour=25
        >>> )
        >>> ax2.set_xlabel('X')
        >>> ax2.set_yticks([])
        >>> ax2.autoscale(tight=True)
        >>> ax2.set_title('Imag component {} Hz'.format(frequency[f_ind]))

        """
        # dxyz = self.vector_distance(xyz)
        # r = repeat_scalar(self.distance(xyz))
        # kr = self.wavenumber*r
        # ikr = 1j*kr

        # front_term = self.moment / (4. * np.pi * r**3) * np.exp(-ikr)
        # symmetric_term = (
        #     repeat_scalar(self.dot_orientation(dxyz)) * dxyz *
        #     (-kr**2 + 3*ikr + 3) / r**2
        # )
        # oriented_term = (
        #     (kr**2 - ikr - 1) *
        #     np.kron(self.orientation, np.ones((dxyz.shape[0], 1)))
        # )

        # return front_term * (symmetric_term + oriented_term)

        n_freq = len(self.frequency)
        n_loc = np.shape(xyz)[0]

        k = self.wavenumber
        r = self.distance(xyz)
        dxyz = self.vector_distance(xyz)

        # (n_freq, n_loc)
        kr = np.outer(k, r)
        ikr = 1j * kr
        tile_r = np.outer(np.ones(n_freq), r)

        front_term = self.moment * (
            1 / (4 * np.pi * tile_r**3) * np.exp(-ikr)
        ).reshape((n_freq, n_loc, 1))
        front_term = np.tile(front_term, (1, 1, 3))

        temp_1 = repeat_scalar(self.dot_orientation(dxyz)) * dxyz
        temp_1 = np.tile(temp_1.reshape((1, n_loc, 3)), (n_freq, 1, 1))
        temp_2 = (-kr**2 + 3*ikr + 3) / tile_r**2
        temp_2 = np.tile(temp_2.reshape((n_freq, n_loc, 1)), (1, 1, 3))
        symmetric_term = temp_1 * temp_2

        temp_1 = (kr**2 - ikr - 1)
        temp_1 = np.tile(temp_1.reshape((n_freq, n_loc, 1)), (1, 1, 3))
        temp_2 = np.kron(self.orientation, np.ones((dxyz.shape[0], 1)))
        temp_2 = np.tile(temp_2.reshape((1, n_loc, 3)), (n_freq, 1, 1))
        oriented_term = temp_1 * temp_2

        return (front_term * (symmetric_term + oriented_term)).squeeze()

    def magnetic_flux_density(self, xyz):
        r"""Magnetic flux density for the harmonic magnetic dipole at a set of gridded locations.

        For a harmonic magnetic dipole oriented in the :math:`\hat{u}` direction with
        moment amplitude :math:`m` and harmonic frequency :math:`f`, this method computes the
        magnetic flux density at the set of gridded xyz locations provided.

        The analytic solution is adapted from Ward and Hohmann (1988). For a harmonic magnetic
        dipole oriented in the :math:`\hat{x}` direction, the solution at vector distance
        :math:`\mathbf{r}` from the dipole is:

        .. math::
            \mathbf{B}(\mathbf{r}) = \frac{\mu m}{4\pi r^3} e^{-ikr} \Bigg [ \Bigg ( \frac{x^2}{r^2}\hat{x} + \frac{xy}{r^2}\hat{y} + \frac{xz}{r^2}\hat{z} \Bigg ) ... \\
            \big ( -k^2r^2 + 3ikr + 3 \big ) \big ( k^2 r^2 - ikr - 1 \big ) \hat{x} \Bigg ]

        where

        .. math::
            k = \sqrt{\omega^2 \mu \varepsilon - i \omega \mu \sigma}

        Parameters
        ----------
        xyz : (n, 3) numpy.ndarray
            Gridded xyz locations

        Returns
        -------
        (n_freq, n_loc, 3) numpy.array of complex
            Magnetic flux density at all frequencies for the gridded
            locations provided. Output array is squeezed when n_freq and/or
            n_loc = 1.

        Examples
        --------
        Here, we define an z-oriented magnetic dipole and plot the magnetic flux density
        on the xz-plane that intercepts y=0.

        >>> from geoana.em.fdem import MagneticDipoleWholeSpace
        >>> from geoana.utils import ndgrid
        >>> from geoana.plotting_utils import plot2Ddata
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt

        Let us begin by defining the electric current dipole.

        >>> frequency = np.logspace(1, 3, 3)
        >>> location = np.r_[0., 0., 0.]
        >>> orientation = np.r_[0., 0., 1.]
        >>> moment = 1.
        >>> sigma = 1.0
        >>> simulation = MagneticDipoleWholeSpace(
        >>>     frequency, location=location, orientation=orientation,
        >>>     moment=moment, sigma=sigma
        >>> )

        Now we create a set of gridded locations and compute the magnetic flux density.

        >>> xyz = ndgrid(np.linspace(-1, 1, 20), np.array([0]), np.linspace(-1, 1, 20))
        >>> B = simulation.magnetic_field(xyz)

        Finally, we plot the real and imaginary components of the magnetic flux density.

        >>> f_ind = 1
        >>> fig = plt.figure(figsize=(6, 3))
        >>> ax1 = fig.add_axes([0.15, 0.15, 0.40, 0.75])
        >>> plot2Ddata(
        >>>     xyz[:, 0::2], np.real(B[f_ind, :, 0::2]), vec=True, ax=ax1, scale='log', ncontour=25
        >>> )
        >>> ax1.set_xlabel('X')
        >>> ax1.set_ylabel('Z')
        >>> ax1.autoscale(tight=True)
        >>> ax1.set_title('Real component {} Hz'.format(frequency[f_ind]))
        >>> ax2 = fig.add_axes([0.6, 0.15, 0.40, 0.75])
        >>> plot2Ddata(
        >>>     xyz[:, 0::2], np.imag(B[f_ind, :, 0::2]), vec=True, ax=ax2, scale='log', ncontour=25
        >>> )
        >>> ax2.set_xlabel('X')
        >>> ax2.set_yticks([])
        >>> ax2.autoscale(tight=True)
        >>> ax2.set_title('Imag component {} Hz'.format(frequency[f_ind]))

        """
        return self.mu * self.magnetic_field(xyz)



"""
======================================================
Static Electromagnetics (:mod:`geoana.em.static`)
======================================================
.. currentmodule:: geoana.em.static

The ``geoana.em.static`` module contains simulation classes for solving
basic electrostatic and magnetostatic problems.


Simulation Classes
==================
.. autosummary::
  :toctree: generated/

  LineCurrentFreeSpace
  MagneticDipoleWholeSpace
  MagneticPoleWholeSpace
  CircularLoopWholeSpace
  ElectrostaticSphere
  MagnetostaticSphere
"""

from geoana.em.static.sphere import ElectrostaticSphere

from geoana.em.static.sphere import MagnetostaticSphere

from geoana.em.static.wholespace import (
    MagneticDipoleWholeSpace,
    MagneticPoleWholeSpace,
    CircularLoopWholeSpace
)

from geoana.em.static.freespace import LineCurrentFreeSpace





# distutils: language=c++
# cython: language_level=3
import numpy as np
cimport numpy as np
cimport cython
from libcpp.complex cimport complex, sqrt, exp
from libcpp cimport bool

ctypedef np.float64_t REAL_t
ctypedef np.intp_t SIZE_t
ctypedef np.complex128_t COMPLEX_t
ctypedef complex[double] complex_t

cdef extern from "_rTE.h" namespace "funcs":
    void rTE(
        complex_t *TE,
        double *frequencies,
        double *lambdas,
        complex_t *sigmas,
        double *mus,
        double *thicks,
        SIZE_t n_frequency,
        SIZE_t n_filter,
        SIZE_t n_layers
        ) nogil

    void rTEgrad(
        complex_t * TE_dsigma,
        complex_t * TE_dmu,
        complex_t * TE_dh,
        double * frequencies,
        double * lambdas,
        complex_t * sigmas,
        double * mus,
        double * h,
        SIZE_t n_frequency,
        SIZE_t n_filter,
        SIZE_t n_layers
    ) nogil

def rTE_forward(frequencies, lamb, sigma, mu, thicknesses):
    """Compute reflection coefficients for Transverse Electric (TE) mode.

    The first layer is considered to be the top most layer. The last
    layer is considered to have infinite thickness. All physical properties 
    are defined starting from the top most layer. 

    Parameters
    ----------
    frequencies : float, numpy.ndarray
        Frequency (Hz); shape = (n_frequency, ).
    lamb : float, numpy.ndarray
        Spatial wavenumber (1/m), shape = (n_filter, ).
    sigma: complex, numpy.ndarray
        Conductivity (S/m), shape = (n_layer, n_frequency).
    mu: complex, numpy.ndarray
        Magnetic permeability (H/m), shape = (n_layer, n_frequency).
    thicknesses: float, numpy.ndarray
        Thickness (m) of each layer, shape = (n_layer-1,).

    Returns
    -------
    rTE: complex numpy.ndarray
        Reflection coefficients, shape = (n_frequency, n_filter)
    """
    # Sigma and mu must be fortran contiguous
    sigma = np.require(sigma, dtype=np.complex128, requirements="F")
    mu = np.require(mu, dtype=np.float64, requirements="F")

    # These just must be contiguous
    lamb = np.require(lamb, dtype=np.float64, requirements="C")
    frequencies = np.require(frequencies, dtype=np.float64, requirements="C")
    thicknesses = np.require(thicknesses, dtype=np.float64, requirements="C")

    # Dimension checking
    if frequencies.shape[0] != sigma.shape[1]:
        raise ValueError(
            f"sigma array's last dimension must be the same as the frequency arrays first. Got {sigma.shape} and {frequencies.shape}"
        )
    if sigma.shape[0] != thicknesses.shape[0]+1:
        raise ValueError(
            f"sigma array's first dimension must match thickness array length+1. Got {sigma.shape[0]} and {thicknesses.shape[0]}"
        )
    if mu.shape != sigma.shape:
        raise ValueError(
            f"mu array must match sigma array shape. Got {mu.shape} and {sigma.shape}"
        )

    cdef:
        REAL_t[:] f = frequencies
        REAL_t[:] lam = lamb
        COMPLEX_t[:, :] sig = sigma
        REAL_t[:, :] c_mu = mu
        REAL_t[:] hs = thicknesses

    cdef:
        SIZE_t n_frequency, n_filter, n_layers


    n_frequency = f.shape[0]
    n_filter = lam.shape[0]
    n_layers = sig.shape[0]

    cdef COMPLEX_t[:, :] out = np.empty((n_frequency, n_filter), dtype=np.complex128, order="F")

    cdef:
        # Types are same size and both pairs of numbers (r, i), so can cast one to the other
        complex_t *out_p = <complex_t *> &out[0, 0]
        complex_t *sig_p = <complex_t *> &sig[0, 0]
        REAL_t *h_p = NULL
    if n_layers > 1:
        h_p = &hs[0]

    rTE(out_p, &f[0], &lam[0], sig_p, &c_mu[0, 0], h_p,
          n_frequency, n_filter, n_layers)

    return np.array(out)

def rTE_gradient(frequencies, lamb, sigma, mu, thicknesses):
    """Compute reflection coefficients for Transverse Electric (TE) mode.

    The first layer is considered to be the top most layer. The last
    layer is considered to have infinite thickness. All physical properties 
    are defined starting from the top most layer. 

    Parameters
    ----------
    frequencies : float, numpy.ndarray
        Frequency (Hz); shape = (n_frequency, ).
    lamb : float, numpy.ndarray
        Spatial wavenumber (1/m), shape = (n_filter, ).
    sigma: complex, numpy.ndarray
        Conductivity (S/m), shape = (n_layer, n_frequency).
    mu: complex, numpy.ndarray
        Magnetic permeability (H/m), shape = (n_layer, n_frequency).
    thicknesses: float, numpy.ndarray
        Thickness (m) of each layer, shape = (n_layer-1,).

    Returns
    -------
    rTE_dsigma: complex numpy.ndarray
        Reflection coefficients gradient w.r.t. conductivity
        shape = (n_layer, n_frequency, n_filter)
    rTE_dh: complex numpy.ndarray
        Reflection coefficients gradient w.r.t. thicknesses
        shape = (n_layer-1, n_frequency, n_filter)
    rTE_dmu: complex numpy.ndarray
        Reflection coefficients gradient w.r.t. magnetic permeability
        shape = (n_layer, n_frequency, n_filter)
    """
    # Require that they are all the same ordering as sigma
    sigma = np.require(sigma, dtype=np.complex128, requirements="F")
    mu = np.require(mu, dtype=np.float64, requirements="F")
    lamb = np.require(lamb, dtype=np.float64, requirements="C")
    frequencies = np.require(frequencies, dtype=np.float64, requirements="C")
    thicknesses = np.require(thicknesses, dtype=np.float64, requirements="C")

    # Dimension checking
    if frequencies.shape[0] != sigma.shape[1]:
        raise ValueError(
            f"sigma array's last dimension must be the same as the frequency arrays first. Got {sigma.shape} and {frequencies.shape}"
        )
    if sigma.shape[0] != thicknesses.shape[0]+1:
        raise ValueError(
            f"sigma array's first dimension must match thickness array length+1. Got {sigma.shape[0]} and {thicknesses.shape[0]}"
        )
    if mu.shape != sigma.shape:
        raise ValueError(
            f"mu array must match sigma array shape. Got {mu.shape} and {sigma.shape}"
        )

    cdef:
        REAL_t[:] f = frequencies
        REAL_t[:] lam = lamb
        COMPLEX_t[:, :] sig = sigma
        REAL_t[:, :] c_mu = mu
        REAL_t[:] hs = thicknesses

    cdef:
        SIZE_t n_frequency, n_filter, n_layers


    n_frequency = f.shape[0]
    n_filter = lam.shape[0]
    n_layers = sig.shape[0]

    cdef COMPLEX_t[:, :, :] gsig = np.empty((n_layers, n_frequency, n_filter), dtype=np.complex128, order="F")
    cdef COMPLEX_t[:, :, :] gh = np.empty((n_layers-1, n_frequency, n_filter), dtype=np.complex128, order="F")
    cdef COMPLEX_t[:, :, :] gmu = np.empty((n_layers, n_frequency, n_filter), dtype=np.complex128, order="F")

    cdef:
        # Types are same size and both pairs of numbers (r, i), so can cast one to the other
        complex_t *sig_p = <complex_t *> &sig[0, 0]
        complex_t *gsig_p = <complex_t *> &gsig[0, 0, 0]
        complex_t *gmu_p = <complex_t *> &gmu[0, 0, 0]
        complex_t *gh_p = NULL
        REAL_t *h_p = NULL
    if n_layers > 1:
        h_p = &hs[0]
        gh_p = <complex_t *> &gh[0, 0, 0]

    rTEgrad(gsig_p, gmu_p, gh_p, &f[0], &lam[0], sig_p, &c_mu[0, 0], h_p,
          n_frequency, n_filter, n_layers)

    return np.array(gsig), np.array(gh), np.array(gmu)



import numpy as np
from geoana.em.base import BaseMagneticDipole
from geoana.em.fdem.base import BaseFDEM, sigma_hat
from scipy.constants import mu_0, epsilon_0
from empymod.utils import check_hankel
from empymod.transform import get_dlf_points
from geoana.kernels.tranverse_electric_reflections import rTE_forward


class MagneticDipoleLayeredHalfSpace(BaseFDEM, BaseMagneticDipole):
    """Simulation class for a harmonic magnetic dipole over a layered halfspace.

    This class is used to simulate the fields produced by a harmonic magnetic dipole
    source over a layered halfspace.

    Parameters
    ----------
    thickness : None or ('*') numpy.ndarray
        Layer thicknesses (m) from the top layer downwards. If ``None`` is entered, you are
        defining a homogeneous halfspace. The bottom layer extends to infinity, thus
        n_layer = n_thickness + 1.
    sigma : (n_layer) np.ndarray or (n_layer, n_frequency) np.ndarray
        Electrical conductivities for all layers (and at all frequencies). For non-dispersive
        conductivity (no chargeability) or for an instance of **MagneticDipoleLayeredHalfSpace**
        at a single frequency, *sigma* is assigned with a (n_layer) np.ndarray.
        For dispersive conductivity and multiple frequencies, *sigma* is assigned with
        a (n_layer, n_frequency) np.ndarray.
    mu : (n_layer) np.ndarray or (n_layer, n_frequency) np.ndarray
        Magnetic permeability for all layers (and at all frequencies). For non-dispersive
        permeability (no viscous remanent magnetization) or for an instance of
        **MagneticDipoleLayeredHalfSpace** at a single frequency, *mu* is assigned with a
        (n_layer) np.ndarray. For dispersive permeability and multiple frequencies, *mu*
        is assigned with a (n_layer, n_frequency) np.ndarray.
    epsilon : (n_layer) np.ndarray or (n_layer, n_frequency) np.ndarray
        Dielectric permittivity for all layers (and at all frequencies). Only applicable when
        *quasistatic* == ``True``. For non-dispersive permittivity or for an instance of
        **MagneticDipoleLayeredHalfSpace** at a single frequency, *epsilon* is assigned with
        a (n_layer) np.ndarray. For dispersive permittivity and multiple frequencies,
        *epsilon* is assigned with a (n_layer, n_frequency) np.ndarray.

    """

    def __init__(self, frequency, thickness, **kwargs):

        self.thickness = thickness
        super().__init__(frequency=frequency, **kwargs)
        self._check_is_valid_location()


    def _check_is_valid_location(self):
        if self.location[2] < 0.0:
            raise ValueError("Source must be above the surface of the earth (i.e. z >= 0.0)")

    @property
    def frequency(self):
        """Frequency (Hz) used for all computations

        Returns
        -------
        numpy.ndarray
            Frequency (or frequencies) in Hz used for all computations
        """
        return self._frequency

    @frequency.setter
    def frequency(self, value):

        # Ensure float or numpy array of float
        try:
            value = np.atleast_1d(value).astype(float)
        except:
            raise TypeError(f"frequencies are not a valid type")

        # Enforce positivity and dimensions
        if (value < 0.).any():
            raise ValueError("All frequencies must be greater than 0")
        if value.ndim > 1:
            raise TypeError(f"frequencies must be ('*') array")

        self._frequency = value

    @property
    def thickness(self):
        """Thicknesses (m) for all layers from top to bottom

        Returns
        -------
        numpy.ndarray
            Thicknesses (m) for all layers from top to bottom
        """
        return self._thickness

    @thickness.setter
    def thickness(self, value):

        # Ensure float or numpy array of float
        try:
            if value is None:
                value = []
            else:
                value = np.atleast_1d(value).astype(float)
        except:
            raise TypeError(f"thickness are not a valid type")

        # Enforce positivity and dimensions
        if (value < 0.).any():
            raise ValueError("Thicknesses must be greater than 0")
        if value.ndim > 1:
            raise TypeError(f"Thicknesses must be ('*') array")

        self._thickness = value


    @property
    def sigma(self):
        """Electrical conductivity for all layers (and frequencies)

        Returns
        -------
        numpy.ndarray (n_layer) or (n_layer, n_frequency)
            Electrical conductivity  for all layers (and frequencies)
        """
        return self._sigma

    @sigma.setter
    def sigma(self, value):

        n_layer = len(self.thickness) + 1
        n_frequency = len(self.frequency)

        # Ensure numpy array of complex
        try:
            if isinstance(value, (list, tuple, np.ndarray)):
                value = np.atleast_1d(value).astype(complex)
            else:
                value = complex(value) * np.ones(n_layer)
        except:
            raise TypeError(f"sigma array is not a valid type")

        # Error if values are non-physical
        if (value.imag < 0.).any():
            raise ValueError("Imaginary components must be >= 0.0")
        if (value.real <= 0.0).any():
            raise ValueError("Real components must be > 0.0")

        # Enforce dimensions
        if (value.ndim == 1) & (len(value) != n_layer):
            raise TypeError(f"sigma must be (n_layer) or (n_layer, n_frequency) np.ndarray")
        elif (value.ndim == 2) & (np.shape(value) != (n_layer, n_frequency)):
            raise TypeError(f"sigma must be (n_layer) or (n_layer, n_frequency) np.ndarray")
        elif value.ndim > 2:
            raise TypeError(f"sigma must be (n_layer) or (n_layer, n_frequency) np.ndarray")

        self._sigma = value

    @property
    def mu(self):
        """Magnetic permeability for all layers (and frequencies)

        Returns
        -------
        numpy.ndarray (n_layer) or (n_layer, n_frequency)
            Magnetic permeability for all layers (and frequencies)
        """
        return self._mu

    @mu.setter
    def mu(self, value):

        n_layer = len(self.thickness) + 1
        n_frequency = len(self.frequency)

        # Ensure float or numpy array of complex
        try:
            if isinstance(value, (list, tuple, np.ndarray)):
                value = np.atleast_1d(value).astype(complex)
            else:
                value = complex(value) * np.ones(n_layer)
        except:
            raise TypeError(f"mu array is not a valid type")

        # Error if values are non-physical
        if (value.imag < 0.).any():
            raise ValueError("Imaginary components must be >= 0")
        if (value.real < mu_0).any():
            raise ValueError("Real components must be >= mu_0")

        # Enforce dimensions
        if (value.ndim == 1) & (len(value) != n_layer):
            raise TypeError(f"mu must be (n_layer) or (n_layer, n_frequency) np.ndarray")
        elif (value.ndim == 2) & (np.shape(value) != (n_layer, n_frequency)):
            raise TypeError(f"mu must be (n_layer) or (n_layer, n_frequency) np.ndarray")
        elif value.ndim > 2:
            raise TypeError(f"mu must be (n_layer) or (n_layer, n_frequency) np.ndarray")

        self._mu = value


    @property
    def epsilon(self):
        """Dielectric permittivity for all layers (and frequencies)

        Returns
        -------
        numpy.ndarray (n_layer) or (n_layer, n_frequency)
            Magnetic permeability for all layers (and frequencies)
        """
        return self._epsilon

    @epsilon.setter
    def epsilon(self, value):

        n_layer = len(self.thickness) + 1
        n_frequency = len(self.frequency)

        # Ensure float or numpy array of complex
        try:
            if isinstance(value, (list, tuple, np.ndarray)):
                value = np.atleast_1d(value).astype(complex)
            else:
                value = complex(value) * np.ones(n_layer)
        except:
            raise TypeError(f"epsilon array is not a valid type")

        # Error if values are non-physical
        if (value.imag < 0.).any():
            raise ValueError("Imaginary components must be >= 0.0")
        if (value.real < epsilon_0).any():
            raise ValueError("Real components must be >= epsilon_0")

        # Enforce dimensions
        if (value.ndim == 1) & (len(value) != n_layer):
            raise TypeError(f"epsilon must be (n_layer) or (n_layer, n_frequency) np.ndarray")
        elif (value.ndim == 2) & (np.shape(value) != (n_layer, n_frequency)):
            raise TypeError(f"epsilon must be (n_layer) or (n_layer, n_frequency) np.ndarray")
        elif value.ndim > 2:
            raise TypeError(f"epsilon must be (n_layer) or (n_layer, n_frequency) np.ndarray")

        self._epsilon = value

    # def _get_valid_properties(self):
    #     thick = self.thickness
    #     n_layer = len(thick)+1
    #     sigma = self.sigma
    #     epsilon = self.epsilon
    #     mu = self.mu
    #     if n_layer != 1:
    #         sigma = self.sigma
    #         if len(sigma) == 1:
    #             sigma = np.ones(n_layer)*sigma
    #         epsilon = self.epsilon
    #         if len(epsilon) == 1:
    #             epsilon = np.ones(n_layer)*epsilon
    #         mu = self.mu
    #         if len(mu) == 1:
    #             mu = np.ones(n_layer)*mu
    #     return thick, sigma, epsilon, mu

    def _get_valid_properties_array(self):

        n_layer = len(self.thickness)+1
        n_frequency = len(self.frequency)

        sigma = self.sigma
        if sigma.ndim == 1:
            sigma = np.tile(sigma.reshape((n_layer, 1)), (1, n_frequency))
        mu = self.mu
        if mu.ndim == 1:
            mu = np.tile(mu.reshape((n_layer, 1)), (1, n_frequency))
        epsilon = self.epsilon
        if epsilon.ndim == 1:
            epsilon = np.tile(epsilon.reshape((n_layer, 1)), (1, n_frequency))

        return self.thickness, sigma, epsilon, mu

    @property
    def sigma_hat(self):
        _, sigma, epsilon, _ = self._get_valid_properties()
        return sigma_hat(
            self.frequency[:, None], sigma, epsilon,
            quasistatic=self.quasistatic
        ).T

    @property
    def wavenumber(self):
        raise NotImplementedError()

    @property
    def skin_depth(self):
        raise NotImplementedError()

    def magnetic_field(self, xyz, field="secondary"):
        r"""
        Compute the magnetic field produced by a magnetic dipole over a layered halfspace.

        Parameters
        ----------
        xyz : numpy.ndarray
            receiver locations of shape (n_locations, 3).
            The z component cannot be below the surface (z >= 0.0).
        field : ("secondary", "total")
            Flag for the type of field to return.

        Returns
        -------
        (n_freq, n_loc, 3) numpy.array of complex
            Magnetic field at all frequencies for the gridded
            locations provided. Output array is squeezed when n_freq and/or
            n_loc = 1.

        Notes
        -----
        We compute the magnetic using the Hankel transform solutions from Ward and Hohmann.
        For the vertical component of the magnetic dipole, the vertical and horizontal fields
        are given by equations 4.45 and 4.46:

        .. math::
            H_\rho = \frac{m_z}{4\pi} \int_0^\infty \bigg [ e^{-u_0 (z - h)} - r_{te} e^{u_0 (z + h)} \bigg ] \lambda^2 J_1 (\lambda \rho) \, d\lambda

        .. math::
            H_z = \frac{m_z}{4\pi} \int_0^\infty \bigg [ e^{-u_0 (z - h)} + r_{te} e^{u_0 (z + h)} \bigg ] \frac{\lambda^3}{u_0} J_0 (\lambda \rho) \, d\lambda

        For the horizontal component of the magnetic dipole, we compute the contribution by adapting
        Ward and Hohmann equations 4.119-4.121; which is for an x-oriented magnetic dipole:

        .. math::
            H_x = & -\frac{m_x}{4\pi} \bigg ( \frac{1}{\rho} - \frac{2x^2}{\rho^3} \bigg ) \int_0^\infty \bigg [ e^{-u_0 (z - h)} - r_{te} e^{u_0 (z + h)} \bigg ] \lambda J_1 (\lambda \rho) \, d\lambda \\
            & -\frac{m_x}{4\pi} \frac{x^2}{\rho^2} \int_0^\infty \bigg [ e^{-u_0 (z - h)} - r_{te} e^{u_0 (z + h)} \bigg ] \lambda^2 J_0 (\lambda \rho) \, d\lambda

        .. math::
            H_y = & \frac{m_x}{2\pi} \frac{xy}{\rho^3} \int_0^\infty \bigg [ e^{-u_0 (z - h)} - r_{te} e^{u_0 (z + h)} \bigg ] \lambda J_1 (\lambda \rho) \, d\lambda \\
            & -\frac{m_x}{4\pi} \frac{xy}{\rho^2} \int_0^\infty \bigg [ e^{-u_0 (z - h)} - r_{te} e^{u_0 (z + h)} \bigg ] \lambda^2 J_0 (\lambda \rho) \, d\lambda

        .. math::
            H_z = \frac{m_x}{4\pi} \frac{x}{\rho}  \int_0^\infty \bigg [ e^{-u_0 (z - h)} + r_{te} e^{u_0 (z + h)} \bigg ] \lambda^2 J_1 (\lambda \rho) \, d\lambda

        Examples
        --------
        Here, we define an z-oriented magnetic dipole at (0, 0, 0) and plot
        the secondary magnetic field at multiple frequencies at (5, 0, 0).
        We compare the secondary fields for a halfspace and for a layered Earth.

        >>> from geoana.em.fdem import (
        >>>     MagneticDipoleHalfSpace, MagneticDipoleLayeredHalfSpace
        >>> )
        >>> from geoana.utils import ndgrid
        >>> from geoana.plotting_utils import plot2Ddata
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt

        Let us begin by defining the electric current dipole.

        >>> frequency = np.logspace(2, 6, 41)
        >>> location = np.r_[0., 0., 0.]
        >>> orientation = np.r_[0., 0., 1.]
        >>> moment = 1.

        We now define the halfspace simulation.

        >>> sigma = 1.0
        >>> simulation_halfspace = MagneticDipoleHalfSpace(
        >>>     frequency, location=location, orientation=orientation,
        >>>     moment=moment, sigma=sigma
        >>> )

        And the layered Earth simulation.

        >>> sigma_top = 0.1
        >>> sigma_middle = 1.0
        >>> sigma_bottom = 0.01
        >>> thickness = np.r_[5., 2.]
        >>> sigma_layers = np.r_[sigma_top, sigma_middle, sigma_bottom]
        >>> simulation_layered = MagneticDipoleLayeredHalfSpace(
        >>>     frequency, thickness, location=location, orientation=orientation,
        >>>     moment=moment, sigma=sigma_layers
        >>> )

        Now we define the receiver location and plot the seconary field.

        >>> xyz = np.c_[5, 0, 0]
        >>> H_halfspace = simulation_halfspace.magnetic_field(xyz, field='secondary')
        >>> H_layered = simulation_layered.magnetic_field(xyz, field='secondary')

        Finally, we plot the real and imaginary components of the magnetic field.

        >>> fig = plt.figure(figsize=(6, 4))
        >>> ax1 = fig.add_axes([0.15, 0.15, 0.8, 0.8])
        >>> ax1.semilogx(frequency, np.real(H_halfspace[:, 2]), 'r', lw=2)
        >>> ax1.semilogx(frequency, np.imag(H_halfspace[:, 2]), 'r--', lw=2)
        >>> ax1.semilogx(frequency, np.real(H_layered[:, 2]), 'b', lw=2)
        >>> ax1.semilogx(frequency, np.imag(H_layered[:, 2]), 'b--', lw=2)
        >>> ax1.set_xlabel('Frequency (Hz)')
        >>> ax1.set_ylabel('Secondary field (H/m)')
        >>> ax1.grid()
        >>> ax1.autoscale(tight=True)
        >>> ax1.legend(['Halfspace: real', 'Halfspace: imag', 'Layered: real', 'Layered: imag'])

        """

        if (xyz[:, 2] < 0.0).any():
            raise ValueError("Cannot compute fields below the surface")
        h = self.location[2]
        dxyz = xyz - self.location
        offsets = np.linalg.norm(dxyz[:, :-1], axis=-1)

        # Comput transform operations
        # -1 gives lagged convolution in dlf
        ht, htarg = check_hankel('dlf', {'dlf': 'key_101_2009', 'pts_per_dec': 0}, 1)
        fhtfilt = htarg['dlf']
        pts_per_dec = htarg['pts_per_dec']

        f = self.frequency
        n_frequency = len(f)

        lambd, int_points = get_dlf_points(fhtfilt, offsets, pts_per_dec)

        thick = self.thickness
        n_layer = len(thick) + 1

        thick, sigma, epsilon, mu = self._get_valid_properties_array()

        # sigh = sigma_hat(
        #     self.frequency[:, None], sigma, epsilon,
        #     quasistatic=self.quasistatic
        # ).T  # this gets sigh with proper shape (n_layer x n_freq) and fortran ordering.
        # mu = np.tile(mu, (n_frequency, 1)).T  # shape(n_layer x n_freq)

        sigh = sigma_hat(
            np.tile(self.frequency.reshape((1, n_frequency)), (n_layer, 1)),
            sigma, epsilon,
            quasistatic=self.quasistatic
        )

        rTE = rTE_forward(f, lambd.reshape(-1), sigh, mu, thick)
        rTE = rTE.reshape((n_frequency, *lambd.shape))

        # secondary is height of receiver plus height of source
        rTE *= np.exp(-lambd*(xyz[:, -1] + h)[:, None])
        # works for variable xyz because each point has it's own lambdas

        src_x, src_y, src_z = self.orientation
        C0x = C0y = C0z = 0.0
        C1x = C1y = C1z = 0.0
        if src_x != 0.0:
            C0x += src_x*(dxyz[:, 0]**2/offsets**2)[:, None]*lambd**2
            C1x += src_x*(1/offsets - 2*dxyz[:, 0]**2/offsets**3)[:, None]*lambd
            C0y += src_x*(dxyz[:, 0]*dxyz[:, 1]/offsets**2)[:, None]*lambd**2
            C1y -= src_x*(2*dxyz[:, 0]*dxyz[:, 1]/offsets**3)[:, None]*lambd
            # C0z += 0.0
            C1z -= (src_x*dxyz[:, 0]/offsets)[:, None]*lambd**2

        if src_y != 0.0:
            C0x += src_y*(dxyz[:, 0]*dxyz[:, 1]/offsets**2)[:, None]*lambd**2
            C1x -= src_y*(2*dxyz[:, 0]*dxyz[:, 1]/offsets**3)[:, None]*lambd
            C0y += src_y*(dxyz[:, 1]**2/offsets**2)[:, None]*lambd**2
            C1y += src_y*(1/offsets - 2*dxyz[:, 1]**2/offsets**3)[:, None]*lambd
            # C0z += 0.0
            C1z -= (src_y*dxyz[:, 1]/offsets)[:, None]*lambd**2

        if src_z != 0.0:
            # C0x += 0.0
            C1x += (src_z*dxyz[:, 0]/offsets)[:, None]*lambd**2
            # C0y += 0.0
            C1y += (src_z*dxyz[:, 1]/offsets)[:, None]*lambd**2
            C0z += src_z*lambd**2
            # C1z += 0.0

        # Do the hankel transform on each component
        em_x = ((C0x*rTE)@fhtfilt.j0 + (C1x*rTE)@fhtfilt.j1)/offsets
        em_y = ((C0y*rTE)@fhtfilt.j0 + (C1y*rTE)@fhtfilt.j1)/offsets
        em_z = ((C0z*rTE)@fhtfilt.j0 + (C1z*rTE)@fhtfilt.j1)/offsets

        if field == "total":
            # add in the primary field
            r = np.linalg.norm(dxyz, axis=-1)
            mdotr = src_x*dxyz[:, 0] + src_y*dxyz[:, 1] + src_z*dxyz[:, 2]

            em_x += 3*dxyz[:, 0]*mdotr/r**5 - src_x/r**3
            em_y += 3*dxyz[:, 1]*mdotr/r**5 - src_y/r**3
            em_z += 3*dxyz[:, 2]*mdotr/r**5 - src_z/r**3

        return self.moment/(4*np.pi)*np.stack((em_x, em_y, em_z), axis=-1).squeeze()
